"""
Given a witness set representation of a pure dimensional solution set,
the functions in this module separate the generic points in the witness set
according to the irreducible components of the solution set.
"""
from ctypes import c_int32, c_double, pointer
from random import uniform
from cmath import exp, pi
from phcpy.version import get_phcfun, int4a2nbr
from phcpy.solutions import make_solution, endmultiplicity
from phcpy.solutions import get_double_solutions
from phcpy.solutions import set_double_solutions
from phcpy.solutions import clear_double_solutions
from phcpy.solutions import get_double_double_solutions
from phcpy.solutions import set_double_double_solutions
from phcpy.solutions import clear_double_double_solutions
from phcpy.solutions import get_quad_double_solutions
from phcpy.solutions import set_quad_double_solutions
from phcpy.solutions import clear_quad_double_solutions
from phcpy.solutions import filter_zero_coordinates
from phcpy.families import cyclic
from phcpy.solver import solve
from phcpy.sets import double_embed, double_double_embed
from phcpy.sets import quad_double_embed
from phcpy.sets import set_double_witness_set
from phcpy.sets import set_double_double_witness_set
from phcpy.sets import set_quad_double_witness_set
from phcpy.sets import set_double_laurent_witness_set
from phcpy.sets import set_double_double_laurent_witness_set
from phcpy.sets import set_quad_double_laurent_witness_set

def set_double_verbose(verbose=True, vrblvl=0):
    """
    Sets the state of the monodromy algorithm in double precision
    to verbose if vrblvl > 0, otherwise it is set to remain mute.
    """
    if vrblvl > 0:
        print('in set_double_verbose, verbose :', verbose)
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> set_double_monodromy_verbose calls phc', end='')
    if verbose:
        retval = phc(630, aaa, bbb, ccc, vrb)
    else:
        retval = phc(39, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def set_double_double_verbose(verbose=True, vrblvl=0):
    """
    Sets the state of the monodromy algorithm in double double precision
    to verbose if vrblvl > 0, otherwise it is set to remain mute.
    """
    if vrblvl > 0:
        print('in set_double_double_verbose, verbose :', verbose)
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> set_double_double_verbose calls phc', end='')
    if verbose:
        retval = phc(660, aaa, bbb, ccc, vrb)
    else:
        retval = phc(658, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def set_quad_double_verbose(verbose=True, vrblvl=0):
    """
    Sets the state of the monodromy algorithm in quad double precision
    to verbose if vrblvl > 0, otherwise it is set to remain mute.
    """
    if vrblvl > 0:
        print('in set_quad_double_verbose, verbose :', verbose)
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> set_quad_double_verbose calls phc', end='')
    if verbose:
        retval = phc(690, aaa, bbb, ccc, vrb)
    else:
        retval = phc(688, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def double_assign_labels(nvr, vrblvl=0):
    """
    Assigns a unique label to the multiplicity field
    for each solution set in double precision.
    """
    if vrblvl > 0:
        print('in double_assign_labels, nvr :', nvr)
    sols = get_double_solutions(vrblvl-1)
    newsols = []
    for (idx, sol) in enumerate(sols):
        data = sol.split('m : 1')
        idxm = f"m : {idx+1}"
        newsol = idxm.join(data)
        newsols.append(newsol)
    clear_double_solutions(vrblvl-1)
    set_double_solutions(nvr, newsols, vrblvl-1)
    return 0

def double_double_assign_labels(nvr, vrblvl=0):
    """
    Assigns a unique label to the multiplicity field
    for each solution set in double double precision.
    """
    if vrblvl > 0:
        print('in double_double_assign_labels, nvr :', nvr)
    sols = get_double_double_solutions(vrblvl-1)
    newsols = []
    for (idx, sol) in enumerate(sols):
        data = sol.split('m : 1')
        idxm = f"m : {idx+1}"
        newsol = idxm.join(data)
        newsols.append(newsol)
    clear_double_double_solutions(vrblvl-1)
    set_double_double_solutions(nvr, newsols, vrblvl-1)
    return 0

def quad_double_assign_labels(nvr, vrblvl=0):
    """
    Assigns a unique label to the multiplicity field
    for each solution set in quad double precision.
    """
    if vrblvl > 0:
        print('in quad_double_assign_labels, nvr :', nvr)
    sols = get_quad_double_solutions(vrblvl-1)
    newsols = []
    for (idx, sol) in enumerate(sols):
        data = sol.split('m : 1')
        idxm = f"m : {idx+1}"
        newsol = idxm.join(data)
        newsols.append(newsol)
    clear_quad_double_solutions(vrblvl-1)
    set_quad_double_solutions(nvr, newsols, vrblvl-1)
    return 0

def initialize_double_sampler(dim, vrblvl=0):
    """
    Initializes the sampling machine with a witness set,
    set in double precision.
    """
    if vrblvl > 0:
        print('in initialize_double_sampler, dim :', dim)
    phc = get_phcfun(vrblvl-1)
    adim = pointer(c_int32(dim))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> initialize_double_sampler calls phc', end='')
    retval = phc(42, adim, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def initialize_double_double_sampler(dim, vrblvl=0):
    """
    Initializes the sampling machine with a witness set,
    set in double double precision.
    """
    if vrblvl > 0:
        print('in initialize_double_double_sampler, dim :', dim)
    phc = get_phcfun(vrblvl-1)
    adim = pointer(c_int32(dim))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> initialize_double_double_sampler calls phc', end='')
    retval = phc(632, adim, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def initialize_quad_double_sampler(dim, vrblvl=0):
    """
    Initializes the sampling machine with a witness set,
    set in quad double precision.
    """
    if vrblvl > 0:
        print('in initialize_quad_double_sampler, dim :', dim)
    phc = get_phcfun(vrblvl-1)
    adim = pointer(c_int32(dim))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> initialize_quad_double_sampler calls phc', end='')
    retval = phc(662, adim, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def initialize_double_laurent_sampler(dim, vrblvl=0):
    """
    Initializes the sampling machine with a witness set,
    for a laurent system set in double precision.
    """
    if vrblvl > 0:
        print('in initialize_double_laurent_sampler, dim :', dim)
    phc = get_phcfun(vrblvl-1)
    adim = pointer(c_int32(dim))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> initialize_double_laurent_sampler calls phc', end='')
    retval = phc(804, adim, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def initialize_double_double_laurent_sampler(dim, vrblvl=0):
    """
    Initializes the sampling machine with a witness set,
    for a laurent system set in double double precision.
    """
    if vrblvl > 0:
        print('in initialize_double_double_laurent_sampler, dim :', dim)
    phc = get_phcfun(vrblvl-1)
    adim = pointer(c_int32(dim))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> initialize_double_double_laurent_sampler calls phc', end='')
    retval = phc(805, adim, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def initialize_quad_double_laurent_sampler(dim, vrblvl=0):
    """
    Initializes the sampling machine with a witness set,
    for a laurent system set in quad double precision.
    """
    if vrblvl > 0:
        print('in initialize_quad_double_laurent_sampler, dim :', dim)
    phc = get_phcfun(vrblvl-1)
    adim = pointer(c_int32(dim))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> initialize_quad_double_laurent_sampler calls phc', end='')
    retval = phc(806, adim, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def initialize_double_monodromy(nbloops, deg, dim, vrblvl=0):
    """
    Initializes to run nbloops monodromy loops to factor
    a positive dimension solution set of dimension dim and of degree deg
    in double precision.
    """
    if vrblvl > 0:
        print('in initialize_double_monodromy', end='')
        print(', nbloops :', nbloops, ', deg :', deg, ', dim :', dim)
    phc = get_phcfun(vrblvl-1)
    anbloops = pointer(c_int32(nbloops))
    bpars = int4a2nbr([deg, dim], vrblvl=vrblvl-1)
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> initialize_double_monodromy calls phc', end='')
    retval = phc(50, anbloops, bpars, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def initialize_double_double_monodromy(nbloops, deg, dim, vrblvl=0):
    """
    Initializes to run nbloops monodromy loops to factor
    a positive dimension solution set of dimension dim and of degree deg
    in double double precision.
    """
    if vrblvl > 0:
        print('in initialize_double_double_monodromy', end='')
        print(', nbloops :', nbloops, ', deg :', deg, ', dim :', dim)
    phc = get_phcfun(vrblvl-1)
    anbloops = pointer(c_int32(nbloops))
    bpars = int4a2nbr([deg, dim], vrblvl=vrblvl-1)
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> initialize_double_double_monodromy calls phc', end='')
    retval = phc(640, anbloops, bpars, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def initialize_quad_double_monodromy(nbloops, deg, dim, vrblvl=0):
    """
    Initializes to run nbloops monodromy loops to factor
    a positive dimension solution set of dimension dim and of degree deg
    in double double precision.
    """
    if vrblvl > 0:
        print('in initialize_quad_double_monodromy', end='')
        print(', nbloops :', nbloops, ', deg :', deg, ', dim :', dim)
    phc = get_phcfun(vrblvl-1)
    anbloops = pointer(c_int32(nbloops))
    bpars = int4a2nbr([deg, dim], vrblvl=vrblvl-1)
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> initialize_double_double_monodromy calls phc', end='')
    retval = phc(670, anbloops, bpars, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def preset_double_solutions(vrblvl=0):
    """
    Initializes the start solutions for monodromy loops to the
    solutions set in double precision.
    """
    if vrblvl > 0:
        print('in preset_double_solutions ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> preset_double_solutions calls phc', end='')
    retval = phc(51, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def preset_double_double_solutions(vrblvl=0):
    """
    Initializes the start solutions for monodromy loops to the
    solutions set in double double precision.
    """
    if vrblvl > 0:
        print('in preset_double_double_solutions ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> preset_double_double_solutions calls phc', end='')
    retval = phc(641, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def preset_quad_double_solutions(vrblvl=0):
    """
    Initializes the start solutions for monodromy loops to the
    solutions set in quad double precision.
    """
    if vrblvl > 0:
        print('in preset_quad_double_solutions ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> preset_quad_double_solutions calls phc', end='')
    retval = phc(671, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def reset_double_solutions(vrblvl=0):
    """
    Resets the start solutions for monodromy loops in double precision
    to the solutions used to initialize the start solutions with.
    """
    if vrblvl > 0:
        print('in reset_double_solutions ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> reset_double_solutions calls phc', end='')
    retval = phc(48, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def reset_double_double_solutions(vrblvl=0):
    """
    Resets the start solutions for monodromy loops in double double precision
    to the solutions used to initialize the start solutions with.
    """
    if vrblvl > 0:
        print('in reset_double_double_solutions ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> reset_double_double_solutions calls phc', end='')
    retval = phc(638, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def reset_quad_double_solutions(vrblvl=0):
    """
    Resets the start solutions for monodromy loops in quad double precision
    to the solutions used to initialize the start solutions with.
    """
    if vrblvl > 0:
        print('in reset_quad_double_solutions ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> reset_quad_double_solutions calls phc', end='')
    retval = phc(668, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def set_double_slice(equ, idx, cff, vrblvl=0):
    """
    Sets the coefficients of slicing equation with index equ
    at position idx to the value in double precision:
    cff[0] + cff[1]*complex(0, 1).
    """
    if vrblvl > 0:
        print('in set_double_slice, equ :', equ, end='')
        print(', in idx :', idx, ', cff :', cff)
    phc = get_phcfun(vrblvl-1)
    aequ = pointer(c_int32(equ))
    bidx = pointer(c_int32(idx))
    cffs = (c_double * 2)()
    cffs[0] = c_double(cff[0])
    cffs[1] = c_double(cff[1])
    ccff = pointer(cffs)
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> set_double_slice calls phc', end='')
    retval = phc(43, aequ, bidx, ccff, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def set_double_double_slice(equ, idx, cff, vrblvl=0):
    """
    Sets the coefficients of slicing equation with index equ
    at position idx to the value in double double precision:
    (cff[0], cff[1]) + (cff[2], cff[3])*complex(0, 1).
    """
    if vrblvl > 0:
        print('in set_double_double_slice, equ :', equ, end='')
        print(', in idx :', idx, ', cff :', cff)
    phc = get_phcfun(vrblvl-1)
    aequ = pointer(c_int32(equ))
    bidx = pointer(c_int32(idx))
    cffs = (c_double * 4)()
    cffs[0] = c_double(cff[0])
    cffs[1] = c_double(cff[1])
    cffs[2] = c_double(cff[2])
    cffs[3] = c_double(cff[3])
    ccff = pointer(cffs)
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> set_double_double_slice calls phc', end='')
    retval = phc(633, aequ, bidx, ccff, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def set_quad_double_slice(equ, idx, cff, vrblvl=0):
    """
    Sets the coefficients of slicing equation with index equ
    at position idx to the value in quad double precision:
    (cff[0], cff[1], cff[2], cff[3])
    + (cff[4], cff[5], cff[6], cff[7])*complex(0, 1).
    """
    if vrblvl > 0:
        print('in set_double_double_slice, equ :', equ, end='')
        print(', in idx :', idx, ', cff :', cff)
    phc = get_phcfun(vrblvl-1)
    aequ = pointer(c_int32(equ))
    bidx = pointer(c_int32(idx))
    cffs = (c_double * 8)()
    cffs[0] = c_double(cff[0])
    cffs[1] = c_double(cff[1])
    cffs[2] = c_double(cff[2])
    cffs[3] = c_double(cff[3])
    cffs[4] = c_double(cff[4])
    cffs[5] = c_double(cff[5])
    cffs[6] = c_double(cff[6])
    cffs[7] = c_double(cff[7])
    ccff = pointer(cffs)
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> set_quad_double_slice calls phc', end='')
    retval = phc(663, aequ, bidx, ccff, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def set_double_trace(first, vrblvl=0):
    """
    Sets the constant coefficient of the first slice,
    for use in the linear trace test in double precision.
    The integer first indicates if it is the first time or not.
    """
    if vrblvl > 0:
        print('in set_double_trace, first :', first)
    if first == 1:
        cff = [-1.0, 0.0]
    else:
        cff = [+1.0, 0.0]
    return set_double_slice(0, 0, cff, vrblvl)

def set_double_double_trace(first, vrblvl=0):
    """
    Sets the constant coefficient of the first slice,
    for use in the linear trace test in double double precision.
    The integer first indicates if it is the first time or not.
    """
    if vrblvl > 0:
        print('in set_double_double_trace, first :', first)
    if first == 1:
        cff = [-1.0, 0.0, 0.0, 0.0]
    else:
        cff = [+1.0, 0.0, 0.0, 0.0]
    return set_double_double_slice(0, 0, cff, vrblvl)

def set_quad_double_trace(first, vrblvl=0):
    """
    Sets the constant coefficient of the first slice,
    for use in the linear trace test in quad double precision.
    The integer first indicates if it is the first time or not.
    """
    if vrblvl > 0:
        print('in set_quad_double_trace, first :', first)
    if first == 1:
        cff = [-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    else:
        cff = [+1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    return set_quad_double_slice(0, 0, cff, vrblvl)

def set_double_gammas(dim, vrblvl=0):
    """
    Sets the gamma constants in double precision
    for the sampler in the monodromy loops.
    Generates as many random complex constants as dim.
    """
    if vrblvl > 0:
        print('in set_double_gammas, dim :', dim)
    phc = get_phcfun(vrblvl-1)
    aidx = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    gamma = (c_double * 2)()
    cgamma = pointer(gamma)
    vrb = c_int32(vrblvl-1)
    retval = 0
    for i in range(dim):
        angle = uniform(0, 2*pi)
        rndgamma = exp(angle*complex(0, 1))
        gamma[0] = c_double(rndgamma.real)
        gamma[1] = c_double(rndgamma.imag)
        if vrblvl > 0:
            print('random gamma :', rndgamma)
        if vrblvl > 0:
            print('-> set_double_gammas calls phc', end='')
        aidx[0] = c_int32(i)
        retval = phc(44, aidx, bbb, cgamma, vrb)
        if vrblvl > 0:
            print(', return value :', retval)
    return retval

def set_double_double_gammas(dim, vrblvl=0):
    """
    Sets the gamma constants in double_double precision
    for the sampler in the monodromy loops.
    Generates as many random complex constants as dim.
    """
    if vrblvl > 0:
        print('in set_double_double_gammas, dim :', dim)
    phc = get_phcfun(vrblvl-1)
    aidx = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    gamma = (c_double * 4)()
    cgamma = pointer(gamma)
    vrb = c_int32(vrblvl-1)
    retval = 0
    for i in range(dim):
        angle = uniform(0, 2*pi)
        rndgamma = exp(angle*complex(0, 1))
        gamma[0] = c_double(rndgamma.real)
        gamma[1] = c_double(0.0)
        gamma[2] = c_double(rndgamma.imag)
        gamma[3] = c_double(0.0)
        if vrblvl > 0:
            print('random gamma :', rndgamma)
        if vrblvl > 0:
            print('-> set_double_double_gammas calls phc', end='')
        aidx[0] = c_int32(i)
        retval = phc(634, aidx, bbb, cgamma, vrb)
        if vrblvl > 0:
            print(', return value :', retval)
    return retval

def set_quad_double_gammas(dim, vrblvl=0):
    """
    Sets the gamma constants in quad_double precision
    for the sampler in the monodromy loops.
    Generates as many random complex constants as dim.
    """
    if vrblvl > 0:
        print('in set_quad_double_gammas, dim :', dim)
    phc = get_phcfun(vrblvl-1)
    aidx = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    gamma = (c_double * 8)()
    cgamma = pointer(gamma)
    vrb = c_int32(vrblvl-1)
    retval = 0
    for i in range(dim):
        angle = uniform(0, 2*pi)
        rndgamma = exp(angle*complex(0, 1))
        gamma[0] = c_double(rndgamma.real)
        gamma[1] = c_double(0.0)
        gamma[2] = c_double(0.0)
        gamma[3] = c_double(0.0)
        gamma[4] = c_double(rndgamma.imag)
        gamma[5] = c_double(0.0)
        gamma[6] = c_double(0.0)
        gamma[7] = c_double(0.0)
        if vrblvl > 0:
            print('random gamma :', rndgamma)
        if vrblvl > 0:
            print('-> set_quad_double_gammas calls phc', end='')
        aidx[0] = c_int32(i)
        retval = phc(664, aidx, bbb, cgamma, vrb)
        if vrblvl > 0:
            print(', return value :', retval)
    return retval

def swap_double_slices(vrblvl=0):
    """
    Swaps the slices with new ones to turn back in one monodromy loop
    in double precision.
    """
    if vrblvl > 0:
        print('in swap_double_slices ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> swap_double_slices calls phc', end='')
    retval = phc(46, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def swap_double_double_slices(vrblvl=0):
    """
    Swaps the slices with new ones to turn back in one monodromy loop
    in double double precision.
    """
    if vrblvl > 0:
        print('in swap_double_double_slices ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> swap_double_double_slices calls phc', end='')
    retval = phc(636, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def swap_quad_double_slices(vrblvl=0):
    """
    Swaps the slices with new ones to turn back in one monodromy loop
    in quad double precision.
    """
    if vrblvl > 0:
        print('in swap_quad_double_slices ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> swap_quad_double_slices calls phc', end='')
    retval = phc(666, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def new_double_slices(dim, nvr, vrblvl=0):
    """
    Generates dim random hyperplanes in double precision
    where nvr is the number of variables.
    """
    if vrblvl > 0:
        print('in new_double_slices, dim :', dim, end='')
        print(', nvr :', nvr)
    fail = 0
    for row in range(dim):
        for col in range(nvr+1):
            angle = uniform(0, 2*pi)
            rancff = exp(angle*complex(0, 1))
            cff = [rancff.real, rancff.imag]
            fail = fail + set_double_slice(row, col, cff, vrblvl)
    return fail

def new_double_double_slices(dim, nvr, vrblvl=0):
    """
    Generates dim random hyperplanes in double double precision
    where nvr is the number of variables.
    """
    if vrblvl > 0:
        print('in new_double_double_slices, dim :', dim, end='')
        print(', nvr :', nvr)
    fail = 0
    for row in range(dim):
        for col in range(nvr+1):
            angle = uniform(0, 2*pi)
            rancff = exp(angle*complex(0, 1))
            cff = [rancff.real, 0.0, rancff.imag, 0.0]
            fail = fail + set_double_double_slice(row, col, cff, vrblvl)
    return fail

def new_quad_double_slices(dim, nvr, vrblvl=0):
    """
    Generates dim random hyperplanes in quad double precision
    where nvr is the number of variables.
    """
    if vrblvl > 0:
        print('in new_quad_double_slices, dim :', dim, end='')
        print(', nvr :', nvr)
    fail = 0
    for row in range(dim):
        for col in range(nvr+1):
            angle = uniform(0, 2*pi)
            rancff = exp(angle*complex(0, 1))
            cff = [rancff.real, 0.0, 0.0, 0.0, rancff.imag, 0.0, 0.0, 0.0]
            fail = fail + set_quad_double_slice(row, col, cff, vrblvl)
    return fail

def double_witness_sample(vrblvl=0):
    """
    Computes a new witness set for a new set of slices,
    in double precision.
    """
    if vrblvl > 0:
        print('in double_witness_sample ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_witness_sample calls phc', end='')
    retval = phc(45, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def double_double_witness_sample(vrblvl=0):
    """
    Computes a new witness set for a new set of slices,
    in double double precision.
    """
    if vrblvl > 0:
        print('in double_double_witness_sample ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_double_witness_sample calls phc', end='')
    retval = phc(635, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def quad_double_witness_sample(vrblvl=0):
    """
    Computes a new witness set for a new set of slices,
    in quad double precision.
    """
    if vrblvl > 0:
        print('in quad_double_witness_sample ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> quad_double_witness_sample calls phc', end='')
    retval = phc(665, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def copy_double_witness_set(vrblvl=0):
    """
    Copies the witness set from the sampler to the system set
    in double precision.
    """
    if vrblvl > 0:
        print('in copy_double_witness_set ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> copy_double_witness_set calls phc', end='')
    retval = phc(47, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def copy_double_double_witness_set(vrblvl=0):
    """
    Copies the witness set from the sampler to the system set
    in double double precision.
    """
    if vrblvl > 0:
        print('in copy_double_double_witness_set ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> copy_double_double_witness_set calls phc', end='')
    retval = phc(637, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def copy_quad_double_witness_set(vrblvl=0):
    """
    Copies the witness set from the sampler to the system set
    in quad double precision.
    """
    if vrblvl > 0:
        print('in copy_quad_double_witness_set ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> copy_quad_double_witness_set calls phc', end='')
    retval = phc(667, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def copy_double_laurent_witness_set(vrblvl=0):
    """
    Copies the witness set from the sampler to the laurent system set
    in double precision.
    """
    if vrblvl > 0:
        print('in copy_double_witness_set ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> copy_double_laurent_witness_set calls phc', end='')
    retval = phc(807, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def copy_double_double_laurent_witness_set(vrblvl=0):
    """
    Copies the witness set from the sampler to the laurent system set
    in double double precision.
    """
    if vrblvl > 0:
        print('in copy_double_double_witness_set ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> copy_double_double_laurent_witness_set calls phc', end='')
    retval = phc(808, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def copy_quad_double_laurent_witness_set(vrblvl=0):
    """
    Copies the witness set from the sampler to the laurent system set
    in quad double precision.
    """
    if vrblvl > 0:
        print('in copy_quad_double_witness_set ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> copy_quad_double_laurent_witness_set calls phc', end='')
    retval = phc(809, aaa, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def double_witness_track(islaurent=False, vrblvl=0):
    """
    Tracks as many paths as set in double precision,
    as many as the size of the witness set.
    If the system is laurent, then islaurent must be True.
    """
    if vrblvl > 0:
        print('in double_witness_track, islaurent :', islaurent)
    double_witness_sample(vrblvl)
    swap_double_slices(vrblvl)
    if islaurent:
        copy_double_laurent_witness_set(vrblvl)
    else:
        copy_double_witness_set(vrblvl)

def double_double_witness_track(islaurent=False, vrblvl=0):
    """
    Tracks as many paths as set in double double precision,
    as many as the size of the witness set.
    If the system is laurent, then islaurent must be True.
    """
    if vrblvl > 0:
        print('in double_double_witness_track, islaurent :', islaurent)
    double_double_witness_sample(vrblvl)
    swap_double_double_slices(vrblvl)
    if islaurent:
        copy_double_double_laurent_witness_set(vrblvl)
    else:
        copy_double_double_witness_set(vrblvl)

def quad_double_witness_track(islaurent=False, vrblvl=0):
    """
    Tracks as many paths as set in quad double precision,
    as many as the size of the witness set.
    If the system is laurent, then islaurent must be True.
    """
    if vrblvl > 0:
        print('in quad_double_witness_track, islaurent :', islaurent)
    quad_double_witness_sample(vrblvl)
    swap_quad_double_slices(vrblvl)
    if islaurent:
        copy_quad_double_laurent_witness_set(vrblvl)
    else:
        copy_quad_double_witness_set(vrblvl)

def double_trace_grid_diagnostics(vrblvl=0):
    """
    Returns the maximal error on the samples in the trace grid
    and the minimal distance between the samples in the trace grid,
    computed in double precision.
    """
    if vrblvl > 0:
        print('in double_trace_grid_diagnostics ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    pars = (c_double * 2)()
    cpars = pointer(pars)
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_trace_grid_diagnostics calls phc', end='')
    retval = phc(56, aaa, bbb, cpars, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    vals = cpars[:2]
    err = vals[0][0]
    dis = vals[0][1]
    if vrblvl > 0:
        print('err :', err, ', dis :', dis)
    return (err, dis)

def double_double_trace_grid_diagnostics(vrblvl=0):
    """
    Returns the maximal error on the samples in the trace grid
    and the minimal distance between the samples in the trace grid,
    computed in double double precision.
    """
    if vrblvl > 0:
        print('in double_double_trace_grid_diagnostics ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    pars = (c_double * 2)()
    cpars = pointer(pars)
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_double_trace_grid_diagnostics calls phc', end='')
    retval = phc(646, aaa, bbb, cpars, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    vals = cpars[:2]
    err = vals[0][0]
    dis = vals[0][1]
    if vrblvl > 0:
        print('err :', err, ', dis :', dis)
    return (err, dis)

def quad_double_trace_grid_diagnostics(vrblvl=0):
    """
    Returns the maximal error on the samples in the trace grid
    and the minimal distance between the samples in the trace grid,
    computed in quad double precision.
    """
    if vrblvl > 0:
        print('in quad_double_trace_grid_diagnostics ...')
    phc = get_phcfun(vrblvl-1)
    aaa = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    pars = (c_double * 2)()
    cpars = pointer(pars)
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> quad_double_trace_grid_diagnostics calls phc', end='')
    retval = phc(676, aaa, bbb, cpars, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    vals = cpars[:2]
    err = vals[0][0]
    dis = vals[0][1]
    if vrblvl > 0:
        print('err :', err, ', dis :', dis)
    return (err, dis)

def double_trace_sum_difference(labels, vrblvl=0):
    """
    Returns the difference between the actual sum at the samples
    defined by the labels to the generic points of a factor
    and the trace sum, in double precision.
    """
    if vrblvl > 0:
        print('in double_trace_sum_difference, labels :', labels)
    phc = get_phcfun(vrblvl-1)
    adim = pointer(c_int32(len(labels)))
    fac = (c_int32 * len(labels))()
    for (idx, label) in enumerate(labels):
        fac[idx] = c_int32(label)
    bfac = pointer(fac)
    cdif = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_trace_sum_difference calls phc', end='')
    retval = phc(57, adim, bfac, cdif, vrb)
    result = cdif[0]
    if vrblvl > 0:
        print(', return value :', retval)
        print('trace sum difference :', result)
    return result

def double_double_trace_sum_difference(labels, vrblvl=0):
    """
    Returns the difference between the actual sum at the samples
    defined by the labels to the generic points of a factor
    and the trace sum, in double double precision.
    """
    if vrblvl > 0:
        print('in double_double_trace_sum_difference, labels :', labels)
    phc = get_phcfun(vrblvl-1)
    adim = pointer(c_int32(len(labels)))
    fac = (c_int32 * len(labels))()
    for (idx, label) in enumerate(labels):
        fac[idx] = c_int32(label)
    bfac = pointer(fac)
    cdif = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_double_trace_sum_difference calls phc', end='')
    retval = phc(647, adim, bfac, cdif, vrb)
    result = cdif[0]
    if vrblvl > 0:
        print(', return value :', retval)
        print('trace sum difference :', result)
    return result

def quad_double_trace_sum_difference(labels, vrblvl=0):
    """
    Returns the difference between the actual sum at the samples
    defined by the labels to the generic points of a factor
    and the trace sum, in quad double precision.
    """
    if vrblvl > 0:
        print('in quad_double_trace_sum_difference, labels :', labels)
    phc = get_phcfun(vrblvl-1)
    adim = pointer(c_int32(len(labels)))
    fac = (c_int32 * len(labels))()
    for (idx, label) in enumerate(labels):
        fac[idx] = c_int32(label)
    bfac = pointer(fac)
    cdif = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> quad_double_trace_sum_difference calls phc', end='')
    retval = phc(677, adim, bfac, cdif, vrb)
    result = cdif[0]
    if vrblvl > 0:
        print(', return value :', retval)
        print('trace sum difference :', result)
    return result

def double_loop_permutation(deg, vrblvl=0):
    """
    Returns the permutation using the solution most recently computed,
    for a set of degree deg, after a loop in double precision.
    """
    if vrblvl > 0:
        print('in double_loop_permutation, deg :', deg)
    phc = get_phcfun(vrblvl-1)
    adeg = pointer(c_int32(deg))
    perm = (c_int32 * deg)()
    bperm = pointer(perm)
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_loop permutation calls phc', end='')
    retval = phc(52, adeg, bperm, ccc, vrb)
    vals = bperm[0:deg]
    result = []
    for idx in range(deg):
        result.append(int(vals[0][idx]))
    if vrblvl > 0:
        print(', return value :', retval)
        print('permutation :', result)
    return result

def double_double_loop_permutation(deg, vrblvl=0):
    """
    Returns the permutation using the solution most recently computed,
    for a set of degree deg, after a loop in double double precision.
    """
    if vrblvl > 0:
        print('in double_double_loop_permutation, deg :', deg)
    phc = get_phcfun(vrblvl-1)
    adeg = pointer(c_int32(deg))
    perm = (c_int32 * deg)()
    bperm = pointer(perm)
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_double_loop permutation calls phc', end='')
    retval = phc(642, adeg, bperm, ccc, vrb)
    vals = bperm[0:deg]
    result = []
    for idx in range(deg):
        result.append(int(vals[0][idx]))
    if vrblvl > 0:
        print(', return value :', retval)
        print('permutation :', result)
    return result

def quad_double_loop_permutation(deg, vrblvl=0):
    """
    Returns the permutation using the solution most recently computed,
    for a set of degree deg, after a loop in quad double precision.
    """
    if vrblvl > 0:
        print('in quad_double_loop_permutation, deg :', deg)
    phc = get_phcfun(vrblvl-1)
    adeg = pointer(c_int32(deg))
    perm = (c_int32 * deg)()
    bperm = pointer(perm)
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> quad_double_loop permutation calls phc', end='')
    retval = phc(672, adeg, bperm, ccc, vrb)
    vals = bperm[0:deg]
    result = []
    for idx in range(deg):
        result.append(int(vals[0][idx]))
    if vrblvl > 0:
        print(', return value :', retval)
        print('permutation :', result)
    return result

def double_factor_count(vrblvl=0):
    """
    Returns the number of factors computed in double precision.
    """
    if vrblvl > 0:
        print('in double_factor_count ...')
    phc = get_phcfun(vrblvl-1)
    anbr = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_factor_count calls phc', end='')
    retval = phc(68, anbr, bbb, ccc, vrb)
    result = anbr[0]
    if vrblvl > 0:
        print(', return value :', retval)
        print('number of factors :', result)
    return result

def double_double_factor_count(vrblvl=0):
    """
    Returns the number of factors computed in double double precision.
    """
    if vrblvl > 0:
        print('in double_double_factor_count ...')
    phc = get_phcfun(vrblvl-1)
    anbr = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_double_factor_count calls phc', end='')
    retval = phc(656, anbr, bbb, ccc, vrb)
    result = anbr[0]
    if vrblvl > 0:
        print(', return value :', retval)
        print('number of factors :', result)
    return result

def quad_double_factor_count(vrblvl=0):
    """
    Returns the number of factors computed in quad double precision.
    """
    if vrblvl > 0:
        print('in quad_double_factor_count ...')
    phc = get_phcfun(vrblvl-1)
    anbr = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> quad_double_factor_count calls phc', end='')
    retval = phc(686, anbr, bbb, ccc, vrb)
    result = anbr[0]
    if vrblvl > 0:
        print(', return value :', retval)
        print('number of factors :', result)
    return result

def update_double_decomposition(deg, perm, vrblvl=0):
    """
    Updates the decomposition with a permutation of deg elements
    computed in double precision.
    Returns the tuple with the previous and the new number of factors.
    """
    if vrblvl > 0:
        print('in update_double_decomposition, deg :', deg)
        print('permutation :', perm)
    phc = get_phcfun(vrblvl-1)
    nbrs = (c_int32 * 2)()
    nbrs[0] = len(perm)
    anbr = pointer(nbrs)
    iperm = (c_int32 * len(perm))()
    for (idx, nbr) in enumerate(perm):
        iperm[idx] = c_int32(nbr)
    bperm = pointer(iperm)
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> update_double_decomposition_count calls phc', end='')
    retval = phc(53, anbr, bperm, ccc, vrb)
    vals = anbr[:2]
    result = (vals[0][0], vals[0][1])
    if vrblvl > 0:
        print(', return value :', retval)
        print('number of factors :', result)
    return result

def update_double_double_decomposition(deg, perm, vrblvl=0):
    """
    Updates the decomposition with a permutation of deg elements
    computed in double double precision.
    Returns the tuple with the previous and the new number of factors.
    """
    if vrblvl > 0:
        print('in update_double_double_decomposition, deg :', deg)
        print('permutation :', perm)
    phc = get_phcfun(vrblvl-1)
    nbrs = (c_int32 * 2)()
    nbrs[0] = len(perm)
    anbr = pointer(nbrs)
    iperm = (c_int32 * len(perm))()
    for (idx, nbr) in enumerate(perm):
        iperm[idx] = c_int32(nbr)
    bperm = pointer(iperm)
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> update_double_double_decomposition_count calls phc', end='')
    retval = phc(643, anbr, bperm, ccc, vrb)
    vals = anbr[:2]
    result = (vals[0][0], vals[0][1])
    if vrblvl > 0:
        print(', return value :', retval)
        print('number of factors :', result)
    return result

def update_quad_double_decomposition(deg, perm, vrblvl=0):
    """
    Updates the decomposition with a permutation of deg elements
    computed in quad double precision.
    Returns the tuple with the previous and the new number of factors.
    """
    if vrblvl > 0:
        print('in update_quad_double_decomposition, deg :', deg)
        print('permutation :', perm)
    phc = get_phcfun(vrblvl-1)
    nbrs = (c_int32 * 2)()
    nbrs[0] = len(perm)
    anbr = pointer(nbrs)
    iperm = (c_int32 * len(perm))()
    for (idx, nbr) in enumerate(perm):
        iperm[idx] = c_int32(nbr)
    bperm = pointer(iperm)
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> update_quad_double_decomposition_count calls phc', end='')
    retval = phc(673, anbr, bperm, ccc, vrb)
    vals = anbr[:2]
    result = (vals[0][0], vals[0][1])
    if vrblvl > 0:
        print(', return value :', retval)
        print('number of factors :', result)
    return result

def double_trace_test(vrblvl=0):
    """
    Runs the trace test on the decompostion in double precision,
    returns True if certified, otherwise returns False.
    """
    if vrblvl > 0:
        print('in double_trace_test ...')
    phc = get_phcfun(vrblvl-1)
    adone = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_trace_test calls phc', end='')
    retval = phc(55, adone, bbb, ccc, vrb)
    result = adone[0] == 1
    if vrblvl > 0:
        print(', return value :', retval)
        print('certified :', result)
    return result

def double_double_trace_test(vrblvl=0):
    """
    Runs the trace test on the decompostion in double double precision,
    returns True if certified, otherwise returns False.
    """
    if vrblvl > 0:
        print('in double_double_trace_test ...')
    phc = get_phcfun(vrblvl-1)
    adone = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_double_trace_test calls phc', end='')
    retval = phc(645, adone, bbb, ccc, vrb)
    result = adone[0] == 1
    if vrblvl > 0:
        print(', return value :', retval)
        print('certified :', result)
    return result

def quad_double_trace_test(vrblvl=0):
    """
    Runs the trace test on the decompostion in quad double precision,
    returns True if certified, otherwise returns False.
    """
    if vrblvl > 0:
        print('in quad_double_trace_test ...')
    phc = get_phcfun(vrblvl-1)
    adone = pointer(c_int32(0))
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> quad_double_trace_test calls phc', end='')
    retval = phc(675, adone, bbb, ccc, vrb)
    result = adone[0] == 1
    if vrblvl > 0:
        print(', return value :', retval)
        print('certified :', result)
    return result

def double_witness_points(idx, deg, vrblvl=0):
    """
    Given an index idx of an irreducible component,
    computed in double precision,
    returns the labels of the witness points that span the component.
    The input deg is the upper bound on the degree of a factor.
    The degree of the factor is the length of the returned list.
    """
    if vrblvl > 0:
        print('in double_witness_points, idx :', idx, end='')
        print(', deg :', deg)
    phc = get_phcfun(vrblvl-1)
    aidx = pointer(c_int32(idx))
    witpts = (c_int32 * deg)()
    bwit = pointer(witpts)
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_witness_points calls phc', end='')
    retval = phc(69, aidx, bwit, ccc, vrb)
    wdeg = aidx[0]
    if vrblvl > 0:
        print(', the return value :', retval)
        print('the degree of factor', idx, ':', wdeg)
    result = []
    vals = bwit[:wdeg]
    for index in range(wdeg):
        result.append(vals[0][index])
    if vrblvl > 0:
        print('labels of witness points :', result)
    return result

def double_double_witness_points(idx, deg, vrblvl=0):
    """
    Given an index idx of an irreducible component,
    computed in double double precision,
    returns the labels of the witness points that span the component.
    The input deg is the upper bound on the degree of a factor.
    The degree of the factor is the length of the returned list.
    """
    if vrblvl > 0:
        print('in double_double_witness_points, idx :', idx, end='')
        print(', deg :', deg)
    phc = get_phcfun(vrblvl-1)
    aidx = pointer(c_int32(idx))
    witpts = (c_int32 * deg)()
    bwit = pointer(witpts)
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_double_witness_points calls phc', end='')
    retval = phc(657, aidx, bwit, ccc, vrb)
    wdeg = aidx[0]
    if vrblvl > 0:
        print(', the return value :', retval)
        print('the degree of factor', idx, ':', wdeg)
    result = []
    vals = bwit[:wdeg]
    for index in range(wdeg):
        result.append(vals[0][index])
    if vrblvl > 0:
        print('labels of witness points :', result)
    return result

def quad_double_witness_points(idx, deg, vrblvl=0):
    """
    Given an index idx of an irreducible component,
    computed in quad double precision,
    returns the labels of the witness points that span the component.
    The input deg is the upper bound on the degree of a factor.
    The degree of the factor is the length of the returned list.
    """
    if vrblvl > 0:
        print('in quad_double_witness_points, idx :', idx, end='')
        print(', deg :', deg)
    phc = get_phcfun(vrblvl-1)
    aidx = pointer(c_int32(idx))
    witpts = (c_int32 * deg)()
    bwit = pointer(witpts)
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> quad_double_witness_points calls phc', end='')
    retval = phc(687, aidx, bwit, ccc, vrb)
    wdeg = aidx[0]
    if vrblvl > 0:
        print(', the return value :', retval)
        print('the degree of factor', idx, ':', wdeg)
    result = []
    vals = bwit[:wdeg]
    for index in range(wdeg):
        result.append(vals[0][index])
    if vrblvl > 0:
        print('labels of witness points :', result)
    return result

def double_decomposition(deg, vrblvl=0):
    """
    Returns the decomposition as a list of labels of witness points
    on the components, computed in double precision.
    """
    if vrblvl > 0:
        print('in double_decomposition, deg :', deg)
    nbr = double_factor_count(vrblvl)
    result = []
    for idx in range(1, nbr+1):
        witpts = double_witness_points(idx, deg, vrblvl)
        trace = double_trace_sum_difference(witpts, vrblvl)
        result.append((witpts, trace))
    return result

def double_double_decomposition(deg, vrblvl=0):
    """
    Returns the decomposition as a list of labels of witness points
    on the components, computed in double double precision.
    """
    if vrblvl > 0:
        print('in double_double_decomposition, deg :', deg)
    nbr = double_double_factor_count(vrblvl)
    result = []
    for idx in range(1, nbr+1):
        witpts = double_double_witness_points(idx, deg, vrblvl)
        trace = double_double_trace_sum_difference(witpts, vrblvl)
        result.append((witpts, trace))
    return result

def quad_double_decomposition(deg, vrblvl=0):
    """
    Returns the decomposition as a list of labels of witness points
    on the components, computed in quad double precision.
    """
    if vrblvl > 0:
        print('in quad_double_decomposition, deg :', deg)
    nbr = quad_double_factor_count(vrblvl)
    result = []
    for idx in range(1, nbr+1):
        witpts = quad_double_witness_points(idx, deg, vrblvl)
        trace = quad_double_trace_sum_difference(witpts, vrblvl)
        result.append((witpts, trace))
    return result

def write_factorization(deco):
    """
    Writes the decomposition in deco.
    """
    for (idx, factor) in enumerate(deco):
        print('  factor', idx+1, ':', factor)

def double_monodromy_breakup(embsys, esols, dim, \
    islaurent=False, verbose=False, nbloops=20, vrblvl=0):
    r"""
    Applies the monodromy breakup algorithm in double precision
    to factor the *dim*-dimensional algebraic set represented by the
    embedded system *embsys* and its solutions *esols*.
    If the embedded polynomial system is a laurent system,
    then *islaurent* must be True.
    If *verbose* is False, then no output is written.
    The value of *nbloops* equals the maximum number of loops.
    """
    if vrblvl > 0:
        print('in double_monodromy_breakup', end='')
        print(', dim :', dim, ', nbloops :', nbloops)
        print(', islaurent :', islaurent, ', verbose :', verbose)
        print('the embedded polynomials :')
        for pol in embsys:
            print(pol)
        print('the generic points :')
        for (idx, sol) in enumerate(esols):
            print('Solution', idx+1, ':')
            print(sol)
    if verbose:
        print('... running monodromy loops in double precision ...')
    deg = len(esols)
    nvr = len(embsys)
    set_double_verbose(verbose, vrblvl)
    if islaurent:
        set_double_laurent_witness_set(nvr, dim, embsys, esols, vrblvl-1)
        double_assign_labels(nvr, vrblvl)
        initialize_double_laurent_sampler(dim, vrblvl)
    else:
        set_double_witness_set(nvr, dim, embsys, esols, vrblvl-1)
        double_assign_labels(nvr, vrblvl)
        initialize_double_sampler(dim, vrblvl)
    initialize_double_monodromy(nbloops, deg, dim, vrblvl)
    preset_double_solutions(vrblvl)
    if verbose:
        print('... initializing the grid for the linear trace ...')
    for i in range(1, 3):
        set_double_trace(i, vrblvl)
        set_double_gammas(nvr, vrblvl)
        double_witness_track(islaurent, vrblvl)
        preset_double_solutions(vrblvl)
        reset_double_solutions(vrblvl)
        swap_double_slices(vrblvl)
    (err, dis) = double_trace_grid_diagnostics(vrblvl)
    if verbose:
        print('The diagnostics of the trace grid :')
        print('  largest error on the samples :', err)
        print('  smallest distance between the samples :', dis)
    for i in range(1, nbloops+1):
        if verbose:
            print(f'... starting loop {i} ...')
        new_double_slices(dim, nvr, vrblvl)
        set_double_gammas(nvr)
        double_witness_track(islaurent, vrblvl)
        clear_double_solutions(vrblvl-1)
        set_double_gammas(nvr)
        double_witness_track(islaurent, vrblvl)
        preset_double_solutions(vrblvl)
        perm = double_loop_permutation(deg, vrblvl)
        if verbose:
            print('new permutation :', perm)
        nb0 = double_factor_count(vrblvl)
        nf0, nf1 = update_double_decomposition(deg, perm, vrblvl)
        if vrblvl > 0:
            print('nf0 :', nf0, ', nf1 :', nf1)
        nb1 = double_factor_count(vrblvl)
        if verbose:
            print(f'number of factors : {nb0} -> {nb1}')
            deco = double_decomposition(deg, vrblvl)
            print('the decomposition :')
            write_factorization(deco)
        done = double_trace_test(vrblvl)
        if done:
            break
        reset_double_solutions(vrblvl)
    return double_decomposition(deg, vrblvl)

def double_double_monodromy_breakup(embsys, esols, dim, \
    islaurent=False, verbose=False, nbloops=20, vrblvl=0):
    r"""
    Applies the monodromy breakup algorithm in double double precision
    to factor the *dim*-dimensional algebraic set represented by the
    embedded system *embsys* and its solutions *esols*.
    If the embedded polynomial system is a laurent system,
    then *islaurent* must be True.
    If *verbose* is False, then no output is written.
    The value of *nbloops* equals the maximum number of loops.
    """
    if vrblvl > 0:
        print('in double_double_monodromy_breakup', end='')
        print(', dim :', dim, ', nbloops :', nbloops)
        print(', islaurent :', islaurent, ', verbose :', verbose)
        print('the embedded polynomials :')
        for pol in embsys:
            print(pol)
        print('the generic points :')
        for (idx, sol) in enumerate(esols):
            print('Solution', idx+1, ':')
            print(sol)
    if verbose:
        print('... running monodromy loops in double double precision ...')
    deg = len(esols)
    nvr = len(embsys)
    set_double_double_verbose(verbose, vrblvl)
    if islaurent:
        set_double_double_laurent_witness_set(nvr, dim, embsys, esols, \
            vrblvl-1)
        double_double_assign_labels(nvr, vrblvl)
        initialize_double_double_laurent_sampler(dim, vrblvl)
    else:
        set_double_double_witness_set(nvr, dim, embsys, esols, vrblvl-1)
        double_double_assign_labels(nvr, vrblvl)
        initialize_double_double_sampler(dim, vrblvl)
    initialize_double_double_monodromy(nbloops, deg, dim, vrblvl)
    preset_double_double_solutions(vrblvl)
    if verbose:
        print('... initializing the grid for the linear trace ...')
    for i in range(1, 3):
        set_double_double_trace(i, vrblvl)
        set_double_double_gammas(nvr, vrblvl)
        double_double_witness_track(islaurent, vrblvl)
        preset_double_double_solutions(vrblvl)
        reset_double_double_solutions(vrblvl)
        swap_double_double_slices(vrblvl)
    (err, dis) = double_double_trace_grid_diagnostics(vrblvl)
    if verbose:
        print('The diagnostics of the trace grid :')
        print('  largest error on the samples :', err)
        print('  smallest distance between the samples :', dis)
    for i in range(1, nbloops+1):
        if verbose:
            print(f'... starting loop {i} ...')
        new_double_double_slices(dim, nvr, vrblvl)
        set_double_double_gammas(nvr)
        double_double_witness_track(islaurent, vrblvl)
        clear_double_double_solutions(vrblvl-1)
        set_double_double_gammas(nvr)
        double_double_witness_track(islaurent, vrblvl)
        preset_double_double_solutions(vrblvl)
        perm = double_double_loop_permutation(deg, vrblvl)
        if verbose:
            print('new permutation :', perm)
        nb0 = double_double_factor_count(vrblvl)
        nf0, nf1 = update_double_double_decomposition(deg, perm, vrblvl)
        if vrblvl > 0:
            print('nf0 :', nf0, ', nf1 :', nf1)
        nb1 = double_double_factor_count(vrblvl)
        if verbose:
            print(f'number of factors : {nb0} -> {nb1}')
            deco = double_double_decomposition(deg, vrblvl)
            print('the decomposition :')
            write_factorization(deco)
        done = double_double_trace_test(vrblvl)
        if done:
            break
        reset_double_double_solutions(vrblvl)
    return double_double_decomposition(deg, vrblvl)

def quad_double_monodromy_breakup(embsys, esols, dim, \
    islaurent=False, verbose=False, nbloops=20, vrblvl=0):
    r"""
    Applies the monodromy breakup algorithm in quad double precision
    to factor the *dim*-dimensional algebraic set represented by the
    embedded system *embsys* and its solutions *esols*.
    If the embedded polynomial system is a laurent system,
    then *islaurent* must be True.
    If *verbose* is False, then no output is written.
    The value of *nbloops* equals the maximum number of loops.
    """
    if vrblvl > 0:
        print('in quad_double_monodromy_breakup', end='')
        print(', dim :', dim, ', nbloops :', nbloops)
        print(', islaurent :', islaurent, ', verbose :', verbose)
        print('the embedded polynomials :')
        for pol in embsys:
            print(pol)
        print('the generic points :')
        for (idx, sol) in enumerate(esols):
            print('Solution', idx+1, ':')
            print(sol)
    if verbose:
        print('... running monodromy loops in quad double precision ...')
    deg = len(esols)
    nvr = len(embsys)
    set_quad_double_verbose(verbose, vrblvl)
    if islaurent:
        set_quad_double_laurent_witness_set(nvr, dim, embsys, esols, \
            vrblvl-1)
        quad_double_assign_labels(nvr, vrblvl)
        initialize_quad_double_laurent_sampler(dim, vrblvl)
    else:
        set_quad_double_witness_set(nvr, dim, embsys, esols, vrblvl-1)
        quad_double_assign_labels(nvr, vrblvl)
        initialize_quad_double_sampler(dim, vrblvl)
    initialize_quad_double_monodromy(nbloops, deg, dim, vrblvl)
    preset_quad_double_solutions(vrblvl)
    if verbose:
        print('... initializing the grid for the linear trace ...')
    for i in range(1, 3):
        set_quad_double_trace(i, vrblvl)
        set_quad_double_gammas(nvr, vrblvl)
        quad_double_witness_track(islaurent, vrblvl)
        preset_quad_double_solutions(vrblvl)
        reset_quad_double_solutions(vrblvl)
        swap_quad_double_slices(vrblvl)
    (err, dis) = quad_double_trace_grid_diagnostics(vrblvl)
    if verbose:
        print('The diagnostics of the trace grid :')
        print('  largest error on the samples :', err)
        print('  smallest distance between the samples :', dis)
    for i in range(1, nbloops+1):
        if verbose:
            print(f'... starting loop {i} ...')
        new_quad_double_slices(dim, nvr, vrblvl)
        set_quad_double_gammas(nvr)
        quad_double_witness_track(islaurent, vrblvl)
        clear_quad_double_solutions(vrblvl-1)
        set_quad_double_gammas(nvr)
        quad_double_witness_track(islaurent, vrblvl)
        preset_quad_double_solutions(vrblvl)
        perm = quad_double_loop_permutation(deg, vrblvl)
        if verbose:
            print('new permutation :', perm)
        nb0 = quad_double_factor_count(vrblvl)
        nf0, nf1 = update_quad_double_decomposition(deg, perm, vrblvl)
        if vrblvl > 0:
            print('nf0 :', nf0, ', nf1 :', nf1)
        nb1 = quad_double_factor_count(vrblvl)
        if verbose:
            print(f'number of factors : {nb0} -> {nb1}')
            deco = quad_double_decomposition(deg, vrblvl)
            print('the decomposition :')
            write_factorization(deco)
        done = quad_double_trace_test(vrblvl)
        if done:
            break
        reset_quad_double_solutions(vrblvl)
    return quad_double_decomposition(deg, vrblvl)

def test_double_assign_labels(vrblvl=0):
    """
    Tests the assigning of labels for solutions
    set in double precision.
    """
    if vrblvl > 0:
        print('in test_double_assign_labels ...')
    names = ['x', 'y']
    sols = []
    for idx in range(3):
        sol = make_solution(names, [idx, 0.0])
        sols.append(sol)
    if vrblvl > 0:
        print('the test solutions :')
        for (idx, sol) in enumerate(sols):
            print('Solution', idx+1, ':')
            print(sol)
    set_double_solutions(2, sols, vrblvl-1)
    double_assign_labels(2, vrblvl)
    newsols = get_double_solutions(vrblvl-1)
    if vrblvl > 0:
        print('after assigning labels :')
    fail = 0
    for (idx, sol) in enumerate(newsols):
        if vrblvl > 0:
            print('Solution', idx+1, ':')
            print(sol)
        (_, mval) = endmultiplicity(sol, vrblvl-1)
        if vrblvl > 0:
            print('mval :', mval, mval == idx+1)
        fail = fail + int(mval != idx+1)
    return fail

def test_double_double_assign_labels(vrblvl=0):
    """
    Tests the assigning of labels for solutions
    set in double double precision.
    """
    if vrblvl > 0:
        print('in test_double_double_assign_labels ...')
    names = ['x', 'y']
    sols = []
    for idx in range(3):
        sol = make_solution(names, [idx, 0.0])
        sols.append(sol)
    if vrblvl > 0:
        print('the test solutions :')
        for (idx, sol) in enumerate(sols):
            print('Solution', idx+1, ':')
            print(sol)
    set_double_double_solutions(2, sols, vrblvl-1)
    double_double_assign_labels(2, vrblvl)
    newsols = get_double_double_solutions(vrblvl-1)
    if vrblvl > 0:
        print('after assigning labels :')
    fail = 0
    for (idx, sol) in enumerate(newsols):
        if vrblvl > 0:
            print('Solution', idx+1, ':')
            print(sol)
        (_, mval) = endmultiplicity(sol, vrblvl-1)
        if vrblvl > 0:
            print('mval :', mval, mval == idx+1)
        fail = fail + int(mval != idx+1)
    return fail

def test_quad_double_assign_labels(vrblvl=0):
    """
    Tests the assigning of labels for solutions
    set in quad double precision.
    """
    if vrblvl > 0:
        print('in test_quad_double_assign_labels ...')
    names = ['x', 'y']
    sols = []
    for idx in range(3):
        sol = make_solution(names, [idx, 0.0])
        sols.append(sol)
    if vrblvl > 0:
        print('the test solutions :')
        for (idx, sol) in enumerate(sols):
            print('Solution', idx+1, ':')
            print(sol)
    set_quad_double_solutions(2, sols, vrblvl-1)
    quad_double_assign_labels(2, vrblvl)
    newsols = get_quad_double_solutions(vrblvl-1)
    if vrblvl > 0:
        print('after assigning labels :')
    fail = 0
    for (idx, sol) in enumerate(newsols):
        if vrblvl > 0:
            print('Solution', idx+1, ':')
            print(sol)
        (_, mval) = endmultiplicity(sol, vrblvl-1)
        if vrblvl > 0:
            print('mval :', mval, mval == idx+1)
        fail = fail + int(mval != idx+1)
    return fail

def test_double_monodromy(vrblvl=0):
    """
    Tests the monodromy breakup in double precision.
    """
    if vrblvl > 0:
        print('in test_double_monodromy ...')
    cyc4 = cyclic(4)
    cyc4e1 = double_embed(4, 1, cyc4, vrblvl=vrblvl-1)
    clear_double_solutions(vrblvl-1)
    c4sols = solve(cyc4e1, vrblvl=vrblvl-1)
    esols = filter_zero_coordinates(c4sols, varname='zz1', tol=1.0e-8, \
        oper='select', vrblvl=vrblvl-1)
    print('the embedded cyclic 4-roots system :')
    for pol in cyc4e1:
        print(pol)
    print('the generic points :')
    for (idx, sol) in enumerate(esols):
        print('Solution', idx+1, ':')
        print(sol)
    fail = int(len(esols) != 4)
    deco = double_monodromy_breakup(cyc4e1, esols, dim=1, \
        verbose=True, nbloops=15, vrblvl=vrblvl)
    if vrblvl > 0:
        print('the decomposition :')
        for (idx, factor) in enumerate(deco):
            print('  factor', idx+1, ':', factor)
    fail = fail + int(len(deco) != 2)
    return fail

def test_double_double_monodromy(vrblvl=0):
    """
    Tests the monodromy breakup in double double precision.
    """
    if vrblvl > 0:
        print('in test_double_double_monodromy ...')
    cyc4 = cyclic(4)
    cyc4e1 = double_double_embed(4, 1, cyc4, vrblvl=vrblvl-1)
    clear_double_double_solutions(vrblvl-1)
    c4sols = solve(cyc4e1, precision='dd', vrblvl=vrblvl-1)
    esols = filter_zero_coordinates(c4sols, varname='zz1', tol=1.0e-8, \
        oper='select', vrblvl=vrblvl-1)
    print('the embedded cyclic 4-roots system :')
    for pol in cyc4e1:
        print(pol)
    print('the generic points :')
    for (idx, sol) in enumerate(esols):
        print('Solution', idx+1, ':')
        print(sol)
    fail = int(len(esols) != 4)
    deco = double_double_monodromy_breakup(cyc4e1, esols, dim=1, \
        verbose=True, nbloops=15, vrblvl=vrblvl)
    if vrblvl > 0:
        print('the decomposition :')
        for (idx, factor) in enumerate(deco):
            print('  factor', idx+1, ':', factor)
    fail = fail + int(len(deco) != 2)
    return fail

def test_quad_double_monodromy(vrblvl=0):
    """
    Tests the monodromy breakup in quad double precision.
    """
    if vrblvl > 0:
        print('in test_quad_double_monodromy ...')
    cyc4 = cyclic(4)
    cyc4e1 = quad_double_embed(4, 1, cyc4, vrblvl=vrblvl-1)
    clear_quad_double_solutions(vrblvl-1)
    c4sols = solve(cyc4e1, precision='qd', vrblvl=vrblvl-1)
    esols = filter_zero_coordinates(c4sols, varname='zz1', tol=1.0e-8, \
        oper='select', vrblvl=vrblvl-1)
    print('the embedded cyclic 4-roots system :')
    for pol in cyc4e1:
        print(pol)
    print('the generic points :')
    for (idx, sol) in enumerate(esols):
        print('Solution', idx+1, ':')
        print(sol)
    fail = int(len(esols) != 4)
    deco = quad_double_monodromy_breakup(cyc4e1, esols, dim=1, \
        verbose=True, nbloops=15, vrblvl=vrblvl)
    if vrblvl > 0:
        print('the decomposition :')
        write_factorization(deco)
    fail = fail + int(len(deco) != 2)
    return fail

def main():
    """
    Runs some tests.
    """
    lvl = 1
    fail = test_double_assign_labels(lvl)
    fail = fail + test_double_double_assign_labels(lvl)
    fail = fail + test_quad_double_assign_labels(lvl)
    fail = fail + test_double_monodromy(lvl)
    fail = fail + test_double_double_monodromy(lvl)
    fail = fail + test_quad_double_monodromy(lvl)
    if fail == 0:
        print('=> All tests passed.')
    else:
        print('Number of failed tests :', fail)

if __name__=='__main__':
    main()



"""
The module sweepers exports the definition of sweep homotopies and
the tracking of solution paths defined by sweep homotopies.
A sweep homotopy is a polynomial system where some of the variables
are considered as parameters.  Given solutions for some parameters
and new values for the parameters, we can track the solution paths
starting at the given solutions and ending at the new solutions for
the new values of the parameters.
The sweep is controlled by a convex linear combination between the
list of start and target values for the parameters.
We distinguish between a complex and a real sweep.
In a complex sweep, with a randomly generated gamma we avoid singularities
along the solution paths, in a complex convex combination between the
start and target values for the parameters.  This complex sweep is
applicable only when the parameter space is convex.
In a real sweep, arc-length parameter continuation is applied.
The algorithms applied in this module are described in the paper by
Kathy Piret and Jan Verschelde: Sweeping Algebraic Curves for Singular 
Solutions.  Journal of Computational and Applied Mathematics,
volume 234, number 4, pages 1228-1237, 2010. 
"""
from ctypes import c_int32, c_double, pointer
from math import sqrt
from phcpy.version import get_phcfun
from phcpy.polynomials import set_double_system
from phcpy.polynomials import set_double_double_system
from phcpy.polynomials import set_quad_double_system
from phcpy.solutions import set_double_solutions, get_double_solutions
from phcpy.solutions import set_double_double_solutions
from phcpy.solutions import get_double_double_solutions
from phcpy.solutions import set_quad_double_solutions
from phcpy.solutions import get_quad_double_solutions
from phcpy.solutions import make_solution, verify

def set_double_start(start, vrblvl=0):
    """
    Sets the values of all parameters to start,
    which contains the consecutive values of all real and imaginary parts
    of the start values of all parameters, in double precision.
    """
    if vrblvl > 0:
        print('in set_double_start, start :')
        for cff in start:
            print(cff)
    size = len(start)
    phc = get_phcfun(vrblvl-1)
    pars = (c_int32 * 2)()
    pars[0] = 0 # double precision
    pars[1] = 0 # start values
    apar = pointer(pars)
    bnum = pointer(c_int32(size))
    vals = (c_double * size)()
    for (idx, coefficient) in enumerate(start):
        vals[idx] = c_double(coefficient)
    pval = pointer(vals)
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> set_double_start calls phc', end='')
    retval = phc(618, apar, bnum, pval, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def set_double_double_start(start, vrblvl=0):
    """
    Sets the values of all parameters to start,
    which contains the consecutive values of all real and imaginary parts
    of the start values of all parameters, in double double precision.
    The length of start must be twice the size of the start
    in double precision.
    """
    if vrblvl > 0:
        print('in set_double_double_start, start :')
        for cff in start:
            print(cff)
    size = len(start)
    phc = get_phcfun(vrblvl-1)
    pars = (c_int32 * 2)()
    pars[0] = 1 # double double precision
    pars[1] = 0 # start values
    apar = pointer(pars)
    bnum = pointer(c_int32(size))
    vals = (c_double * size)()
    for (idx, coefficient) in enumerate(start):
        vals[idx] = c_double(coefficient)
    pval = pointer(vals)
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> set_double_double_start calls phc', end='')
    retval = phc(618, apar, bnum, pval, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def set_quad_double_start(start, vrblvl=0):
    """
    Sets the values of all parameters to start,
    which contains the consecutive values of all real and imaginary parts
    of the start values of all parameters, in quad double precision.
    The length of start must be four times the size of the start
    in double precision.
    """
    if vrblvl > 0:
        print('in set_quad_double_start, start :')
        for cff in start:
            print(cff)
    size = len(start)
    phc = get_phcfun(vrblvl-1)
    pars = (c_int32 * 2)()
    pars[0] = 2 # quad double precision
    pars[1] = 0 # start values
    apar = pointer(pars)
    bnum = pointer(c_int32(size))
    vals = (c_double * size)()
    for (idx, coefficient) in enumerate(start):
        vals[idx] = c_double(coefficient)
    pval = pointer(vals)
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> set_quad_double_start calls phc', end='')
    retval = phc(618, apar, bnum, pval, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def set_double_target(target, vrblvl=0):
    """
    Sets the values of all parameters to target,
    which contains the consecutive values of all real and imaginary parts
    of the target values of all parameters, in double precision.
    """
    if vrblvl > 0:
        print('in set_double_target, target :')
        for cff in target:
            print(cff)
    size = len(target)
    phc = get_phcfun(vrblvl-1)
    pars = (c_int32 * 2)()
    pars[0] = 0 # double precision
    pars[1] = 1 # target values
    apar = pointer(pars)
    bnum = pointer(c_int32(size))
    vals = (c_double * size)()
    for (idx, coefficient) in enumerate(target):
        vals[idx] = c_double(coefficient)
    pval = pointer(vals)
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> set_double_target calls phc', end='')
    retval = phc(618, apar, bnum, pval, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def set_double_double_target(target, vrblvl=0):
    """
    Sets the values of all parameters to target,
    which contains the consecutive values of all real and imaginary parts
    of the target values of all parameters, in double double precision.
    The length of start must be twice the size of the start
    in double precision.
    """
    if vrblvl > 0:
        print('in set_double_double_target, target :')
        for cff in target:
            print(cff)
    size = len(target)
    phc = get_phcfun(vrblvl-1)
    pars = (c_int32 * 2)()
    pars[0] = 1 # double double precision
    pars[1] = 1 # target values
    apar = pointer(pars)
    bnum = pointer(c_int32(size))
    vals = (c_double * size)()
    for (idx, coefficient) in enumerate(target):
        vals[idx] = c_double(coefficient)
    pval = pointer(vals)
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> set_double_double_target calls phc', end='')
    retval = phc(618, apar, bnum, pval, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def set_quad_double_target(target, vrblvl=0):
    """
    Sets the values of all parameters to target,
    which contains the consecutive values of all real and imaginary parts
    of the target values of all parameters, in double double precision.
    The length of start must be four times the size of the start
    in double precision.
    """
    if vrblvl > 0:
        print('in set_quad_double_target, target :')
        for cff in target:
            print(cff)
    size = len(target)
    phc = get_phcfun(vrblvl-1)
    pars = (c_int32 * 2)()
    pars[0] = 2 # quad double precision
    pars[1] = 1 # target values
    apar = pointer(pars)
    bnum = pointer(c_int32(size))
    vals = (c_double * size)()
    for (idx, coefficient) in enumerate(target):
        vals[idx] = c_double(coefficient)
    pval = pointer(vals)
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> set_quad_double_target calls phc', end='')
    retval = phc(618, apar, bnum, pval, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def set_parameter_names(neq, nvr, pars, vrblvl=0):
    """
    Defines which variables serve as parameters,
    by providing the names of the parameters in the list pars.
    The number of equations is given in neq and
    the number of variables is in nvr.
    """
    if vrblvl > 0:
        print('in set_parameter_names, neq :', neq, end='')
        print(', nvr :', nvr)
        print('parameters :', pars)
    phc = get_phcfun(vrblvl-1)
    apars = (c_int32 * 4)()
    apars[0] = neq
    apars[1] = nvr
    apars[2] = len(pars)
    parnames = ' '.join(pars)
    nbc = len(parnames)
    apars[3] = nbc
    apar = pointer(apars)
    size = nbc
    names = (c_int32 *size)()
    for (idx, letter) in enumerate(parnames):
        names[idx] = c_int32(ord(letter))
    bnames = pointer(names)
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> set_parameter_names calls phc', end='')
    retval = phc(611, apar, bnames, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def double_complex_sweep_run(gchoice, regamma, imgamma, vrblvl=0):
    """
    Starts the trackers in a complex convex parameter homotopy,
    in double precision, where the indices to the parameters,
    start and target values are already defined,
    and the systems and solution are set in double precision.
    The input parameter gchoice is 0, 1, or 2, for respectively
    a randomly generated gamma (0), or no gamma (1), or a user given
    gamma with real and imaginary parts in regamma and imgamma.
    With a random gamma, this is known as cheater's homotopy.
    """
    if vrblvl > 0:
        print('in double_complex_sweep_run, gchoice :', gchoice)
        print('regamma :', regamma, '  imgamma :', imgamma)
    phc = get_phcfun(vrblvl-1)
    pars = (c_int32 * 2)()
    pars[0] = 0  # double precision
    pars[1] = gchoice
    apar = pointer(pars)
    bbb = pointer(c_int32(0))
    gamma = (c_double * 2)()
    gamma[0] = regamma
    gamma[1] = imgamma
    cgamma = pointer(gamma)
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_complex_sweep_run calls phc', end='')
    retval = phc(620, apar, bbb, cgamma, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def double_double_complex_sweep_run(gchoice, regamma, imgamma, vrblvl=0):
    """
    Starts the trackers in a complex convex parameter homotopy,
    in double double precision, where the indices to the parameters,
    start and target values are already defined,
    and the systems and solution are set in double double precision.
    The input parameter gchoice is 0, 1, or 2, for respectively
    a randomly generated gamma (0), or no gamma (1), or a user given
    gamma with real and imaginary parts in regamma and imgamma.
    With a random gamma, this is known as cheater's homotopy.
    """
    if vrblvl > 0:
        print('in double_double_complex_sweep_run, gchoice :', gchoice)
        print('regamma :', regamma, '  imgamma :', imgamma)
    phc = get_phcfun(vrblvl-1)
    pars = (c_int32 * 2)()
    pars[0] = 1  # double double precision
    pars[1] = gchoice
    apar = pointer(pars)
    bbb = pointer(c_int32(0))
    gamma = (c_double * 4)()
    gamma[0] = regamma
    gamma[1] = 0.0
    gamma[2] = imgamma
    gamma[3] = 0.0
    cgamma = pointer(gamma)
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_double_complex_sweep_run calls phc', end='')
    retval = phc(620, apar, bbb, cgamma, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def quad_double_complex_sweep_run(gchoice, regamma, imgamma, vrblvl=0):
    """
    Starts the trackers in a complex convex parameter homotopy,
    in quad double precision, where the indices to the parameters,
    start and target values are already defined,
    and the systems and solution are set in double double precision.
    The input parameter gchoice is 0, 1, or 2, for respectively
    a randomly generated gamma (0), or no gamma (1), or a user given
    gamma with real and imaginary parts in regamma and imgamma.
    With a random gamma, this is known as cheater's homotopy.
    """
    if vrblvl > 0:
        print('in quad_double_complex_sweep_run, gchoice :', gchoice)
        print('regamma :', regamma, '  imgamma :', imgamma)
    phc = get_phcfun(vrblvl-1)
    pars = (c_int32 * 2)()
    pars[0] = 2  # quad double precision
    pars[1] = gchoice
    apar = pointer(pars)
    bbb = pointer(c_int32(0))
    gamma = (c_double * 8)()
    gamma[0] = regamma
    gamma[1] = 0.0
    gamma[2] = 0.0
    gamma[3] = 0.0
    gamma[4] = imgamma
    gamma[5] = 0.0
    gamma[6] = 0.0
    gamma[7] = 0.0
    cgamma = pointer(gamma)
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> quad_double_complex_sweep_run calls phc', end='')
    retval = phc(620, apar, bbb, cgamma, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def double_real_sweep_run(vrblvl=0):
    """
    Starts a sweep with a natural parameter in a family of n equations
    in n+1 variables, where the last variable is the artificial parameter s
    that moves the one natural parameter from a start to target value.
    The last equation is of the form (1-s)*(A - v[0]) + s*(A - v[1]),
    where A is the natural parameter, going from the start value v[0]
    to the target value v[1], already set in double precision.
    Also set before calling the function are the homotopy and the start
    solutions, where every solution has the value v[0] for the A variable.
    The sweep stops when s reaches the value v[1],
    or when a singularity is encountered on the path.
    """
    if vrblvl > 0:
        print('in double_real_sweep ...')
    phc = get_phcfun(vrblvl-1)
    aprc = pointer(c_int32(0)) # double precision
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_real_sweep_run calls phc', end='')
    retval = phc(621, aprc, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def double_double_real_sweep_run(vrblvl=0):
    """
    Starts a sweep with a natural parameter in a family of n equations
    in n+1 variables, where the last variable is the artificial parameter s
    that moves the one natural parameter from a start to target value.
    The last equation is of the form (1-s)*(A - v[0]) + s*(A - v[1]),
    where A is the natural parameter, going from the start value v[0]
    to the target value v[1], already set in double double precision.
    Also set before calling the function are the homotopy and the start
    solutions, where every solution has the value v[0] for the A variable.
    The sweep stops when s reaches the value v[1],
    or when a singularity is encountered on the path.
    """
    if vrblvl > 0:
        print('in double_double_real_sweep ...')
    phc = get_phcfun(vrblvl-1)
    aprc = pointer(c_int32(1)) # double double precision
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> double_double_real_sweep_run calls phc', end='')
    retval = phc(621, aprc, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def quad_double_real_sweep_run(vrblvl=0):
    """
    Starts a sweep with a natural parameter in a family of n equations
    in n+1 variables, where the last variable is the artificial parameter s
    that moves the one natural parameter from a start to target value.
    The last equation is of the form (1-s)*(A - v[0]) + s*(A - v[1]),
    where A is the natural parameter, going from the start value v[0]
    to the target value v[1], already set in quad double precision.
    Also set before calling the function are the homotopy and the start
    solutions, where every solution has the value v[0] for the A variable.
    The sweep stops when s reaches the value v[1],
    or when a singularity is encountered on the path.
    """
    if vrblvl > 0:
        print('in quad_double_real_sweep ...')
    phc = get_phcfun(vrblvl-1)
    aprc = pointer(c_int32(2)) # quad double precision
    bbb = pointer(c_int32(0))
    ccc = pointer(c_double(0.0))
    vrb = c_int32(vrblvl-1)
    if vrblvl > 0:
        print('-> quad_double_real_sweep_run calls phc', end='')
    retval = phc(621, aprc, bbb, ccc, vrb)
    if vrblvl > 0:
        print(', return value :', retval)
    return retval

def double_complex_sweep(pols, sols, nvar, pars, start, target, vrblvl=0):
    r"""
    For the polynomials in the list of strings *pols*
    and the solutions in *sols* for the values in the list *start*,
    a sweep through the parameter space will be performed
    in double precision to the target values of
    the parameters in the list *target*.
    The number of variables in the polynomials and the solutions
    must be the same and be equal to the value of *nvar*.
    The list of symbols in *pars* contains the names of the variables
    in the polynomials *pols* that serve as parameters.
    The size of the lists *pars*, *start*, and *target* must be same.
    """
    if vrblvl > 0:
        print('in double_complex_sweep, nvar :', nvar)
        print('the polynomials :')
        for pol in pols:
            print(pol)
        print('the solutions :')
        for (idx, sol) in enumerate(sols):
            print('Solution', idx+1, ':')
            print(sol)
        print('parameters :', pars)
        print('start :', start, '  target :', target)
    set_double_system(nvar, pols, vrblvl-1)
    set_double_solutions(nvar, sols, vrblvl-1)
    set_parameter_names(len(pols), nvar, pars, vrblvl)
    set_double_start(start, vrblvl)
    set_double_target(target, vrblvl)
    double_complex_sweep_run(0, 0.0, 0.0)
    result = get_double_solutions(vrblvl-1)
    return result

def double_double_complex_sweep(pols, sols, nvar, pars, start, target, \
    vrblvl=0):
    r"""
    For the polynomials in the list of strings *pols*
    and the solutions in *sols* for the values in the list *start*,
    a sweep through the parameter space will be performed
    in double double precision to the target values of
    the parameters in the list *target*.
    The number of variables in the polynomials and the solutions
    must be the same and be equal to the value of *nvar*.
    The list of symbols in *pars* contains the names of the variables
    in the polynomials *pols* that serve as parameters.
    The size of the lists *pars*, *start*, and *target* must be same.
    """
    if vrblvl > 0:
        print('in double_double_complex_sweep, nvar :', nvar)
        print('the polynomials :')
        for pol in pols:
            print(pol)
        print('the solutions :')
        for (idx, sol) in enumerate(sols):
            print('Solution', idx+1, ':')
            print(sol)
        print('parameters :', pars)
        print('start :', start, '  target :', target)
    set_double_double_system(nvar, pols, vrblvl-1)
    set_double_double_solutions(nvar, sols, vrblvl-1)
    set_parameter_names(len(pols), nvar, pars, vrblvl)
    set_double_double_start(start, vrblvl)
    set_double_double_target(target, vrblvl)
    double_double_complex_sweep_run(0, 0.0, 0.0)
    result = get_double_double_solutions(vrblvl-1)
    return result

def quad_double_complex_sweep(pols, sols, nvar, pars, start, target, \
    vrblvl=0):
    r"""
    For the polynomials in the list of strings *pols*
    and the solutions in *sols* for the values in the list *start*,
    a sweep through the parameter space will be performed
    in double double precision to the target values of
    the parameters in the list *target*.
    The number of variables in the polynomials and the solutions
    must be the same and be equal to the value of *nvar*.
    The list of symbols in *pars* contains the names of the variables
    in the polynomials *pols* that serve as parameters.
    The size of the lists *pars*, *start*, and *target* must be same.
    """
    if vrblvl > 0:
        print('in quad_double_complex_sweep, nvar :', nvar)
        print('the polynomials :')
        for pol in pols:
            print(pol)
        print('the solutions :')
        for (idx, sol) in enumerate(sols):
            print('Solution', idx+1, ':')
            print(sol)
        print('parameters :', pars)
        print('start :', start, '  target :', target)
    set_quad_double_system(nvar, pols, vrblvl-1)
    set_quad_double_solutions(nvar, sols, vrblvl-1)
    set_parameter_names(len(pols), nvar, pars, vrblvl)
    set_quad_double_start(start, vrblvl)
    set_quad_double_target(target, vrblvl)
    quad_double_complex_sweep_run(0, 0.0, 0.0)
    result = get_quad_double_solutions(vrblvl-1)
    return result

def double_real_sweep(pols, sols, par='s', start=0.0, target=1.0, vrblvl=0):
    r"""
    A real sweep homotopy is a family of n equations in n+1 variables,
    where one of the variables is the artificial parameter s which moves
    from 0.0 to 1.0.  The last equation can then be of the form

    (1 - s)*(lambda - L[0]) + s*(lambda - L[1]) = 0 so that,

    at s = 0, the natural parameter lambda has the value L[0], and

    at s = 1, the natural parameter lambda has the value L[1].

    Thus: as s moves from 0 to 1, lambda goes from L[0] to L[1].

    All solutions in the list *sols* must have then the value L[0]
    for the variable lambda.
    The sweep stops when the target value for s is reached
    or when a singular solution is encountered.
    Computations happen in double precision.
    """
    if vrblvl > 0:
        print('in double_real_sweep ...')
        print('the polynomials :')
        for pol in pols:
            print(pol)
        print('the solutions :')
        for (idx, sol) in enumerate(sols):
            print('Solution', idx+1, ':')
            print(sol)
        print('parameters :', par)
        print('start :', start, '  target :', target)
    nvar = len(pols) + 1
    set_double_system(nvar, pols, vrblvl-1)
    set_double_solutions(nvar, sols, vrblvl-1)
    set_parameter_names(len(pols), nvar, [par], vrblvl)
    set_double_start([start], vrblvl)
    set_double_target([target], vrblvl)
    double_real_sweep_run(vrblvl)
    result = get_double_solutions(vrblvl-1)
    return result

def double_double_real_sweep(pols, sols, par='s', start=0.0, target=1.0, \
    vrblvl=0):
    r"""
    A real sweep homotopy is a family of n equations in n+1 variables,
    where one of the variables is the artificial parameter s which moves
    from 0.0 to 1.0.  The last equation can then be of the form

    (1 - s)*(lambda - L[0]) + s*(lambda - L[1]) = 0 so that,

    at s = 0, the natural parameter lambda has the value L[0], and

    at s = 1, the natural parameter lambda has the value L[1].

    Thus: as s moves from 0 to 1, lambda goes from L[0] to L[1].

    All solutions in the list *sols* must have then the value L[0]
    for the variable lambda.
    The sweep stops when the target value for s is reached
    or when a singular solution is encountered.
    Computations happen in double double precision.
    """
    if vrblvl > 0:
        print('in double_double_real_sweep ...')
        print('the polynomials :')
        for pol in pols:
            print(pol)
        print('the solutions :')
        for (idx, sol) in enumerate(sols):
            print('Solution', idx+1, ':')
            print(sol)
        print('parameters :', par)
        print('start :', start, '  target :', target)
    nvar = len(pols) + 1
    set_double_double_system(nvar, pols, vrblvl-1)
    set_double_double_solutions(nvar, sols, vrblvl-1)
    set_parameter_names(len(pols), nvar, [par], vrblvl)
    set_double_double_start([start], vrblvl)
    set_double_double_target([target], vrblvl)
    double_double_real_sweep_run(vrblvl)
    result = get_double_double_solutions(vrblvl-1)
    return result

def quad_double_real_sweep(pols, sols, par='s', start=0.0, target=1.0, \
    vrblvl=0):
    r"""
    A real sweep homotopy is a family of n equations in n+1 variables,
    where one of the variables is the artificial parameter s which moves
    from 0.0 to 1.0.  The last equation can then be of the form

    (1 - s)*(lambda - L[0]) + s*(lambda - L[1]) = 0 so that,

    at s = 0, the natural parameter lambda has the value L[0], and

    at s = 1, the natural parameter lambda has the value L[1].

    Thus: as s moves from 0 to 1, lambda goes from L[0] to L[1].

    All solutions in the list *sols* must have then the value L[0]
    for the variable lambda.
    The sweep stops when the target value for s is reached
    or when a singular solution is encountered.
    Computations happen in quad double precision.
    """
    if vrblvl > 0:
        print('in quad_double_real_sweep ...')
        print('the polynomials :')
        for pol in pols:
            print(pol)
        print('the solutions :')
        for (idx, sol) in enumerate(sols):
            print('Solution', idx+1, ':')
            print(sol)
        print('parameters :', par)
        print('start :', start, '  target :', target)
    nvar = len(pols) + 1
    set_quad_double_system(nvar, pols, vrblvl-1)
    set_quad_double_solutions(nvar, sols, vrblvl-1)
    set_parameter_names(len(pols), nvar, [par], vrblvl)
    set_quad_double_start([start], vrblvl)
    set_quad_double_target([target], vrblvl)
    quad_double_real_sweep_run(vrblvl)
    result = get_quad_double_solutions(vrblvl-1)
    return result

def test_double_complex_sweep(vrblvl=0):
    """
    Runs a complex sweep on two points on the unit circle.
    Although we start at two points with real coordinates
    and we end at two points that have nonzero imaginary parts,
    the sweep does not encounter a singularity because of
    the random complex gamma constant.
    """
    if vrblvl > 0:
        print('in test_double_complex_sweep ...')
    circle = ['x^2 + y^2 - 1;']
    first = make_solution(['x', 'y'], [0, 1])
    second = make_solution(['x', 'y'], [0, -1])
    startsols = [first, second]
    xpar = ['x']
    start = [0, 0]  # real and imaginary parts of the start value
    target = [2, 0]
    newsols = double_complex_sweep\
                 (circle, startsols, 2, xpar, start, target, vrblvl)
    if vrblvl > 0:
        for (idx, sol) in enumerate(newsols):
            print('Solution', idx+1, ':')
            print(sol)
    err = verify(circle, newsols, vrblvl-1)
    if vrblvl > 0:
        print('the error :', err)
    fail = int(err > 1.0e-8)
    return fail

def test_double_double_complex_sweep(vrblvl=0):
    """
    Runs a complex sweep on two points on the unit circle,
    in double double precision.
    Although we start at two points with real coordinates
    and we end at two points that have nonzero imaginary parts,
    the sweep does not encounter a singularity because of
    the random complex gamma constant.
    """
    if vrblvl > 0:
        print('in test_double_double_complex_sweep ...')
    circle = ['x^2 + y^2 - 1;']
    first = make_solution(['x', 'y'], [0, 1])
    second = make_solution(['x', 'y'], [0, -1])
    startsols = [first, second]
    xpar = ['x']
    # real and imaginary parts of the start value
    start = [0.0, 0.0, 0.0, 0.0] # as double doubles
    target = [2.0, 0.0, 0.0, 0.0]
    newsols = double_double_complex_sweep\
                 (circle, startsols, 2, xpar, start, target, vrblvl)
    if vrblvl > 0:
        for (idx, sol) in enumerate(newsols):
            print('Solution', idx+1, ':')
            print(sol)
    err = verify(circle, newsols, vrblvl-1)
    if vrblvl > 0:
        print('the error :', err)
    fail = int(err > 1.0e-8)
    return fail

def test_quad_double_complex_sweep(vrblvl=0):
    """
    Runs a complex sweep on two points on the unit circle,
    in quad double precision.
    Although we start at two points with real coordinates
    and we end at two points that have nonzero imaginary parts,
    the sweep does not encounter a singularity because of
    the random complex gamma constant.
    """
    if vrblvl > 0:
        print('in test_quad_double_complex_sweep ...')
    circle = ['x^2 + y^2 - 1;']
    first = make_solution(['x', 'y'], [0, 1])
    second = make_solution(['x', 'y'], [0, -1])
    startsols = [first, second]
    xpar = ['x']
    # real and imaginary parts of the start value
    start = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0] # as quad doubles
    target = [2.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    newsols = quad_double_complex_sweep\
                 (circle, startsols, 2, xpar, start, target, vrblvl)
    if vrblvl > 0:
        for (idx, sol) in enumerate(newsols):
            print('Solution', idx+1, ':')
            print(sol)
    err = verify(circle, newsols, vrblvl-1)
    if vrblvl > 0:
        print('the error :', err)
    fail = int(err > 1.0e-8)
    return fail

def test_double_real_sweep(vrblvl=0):
    """
    Runs a real sweep on two points on the unit circle: (1,0), (-1,0),
    moving the second coordinate from 0 to 2.
    The sweep will stop at the quadratic turning point: (0,1).
    We can also run the sweep starting at two complex points:
    (2*j, sqrt(5)) and (-2*j, sqrt(5)), moving the second coordinate
    from sqrt(5) to 0.  This sweep will also stop at (0,1).
    """
    if vrblvl > 0:
        print('in test_double_real_sweep ...')
    circle = ['x^2 + y^2 - 1;', 'y*(1-s) + (y-2)*s;']
    first = make_solution(['x', 'y', 's'], [1, 0, 0])
    second = make_solution(['x', 'y', 's'], [-1, 0, 0])
    startsols = [first, second]
    newsols = double_real_sweep(circle, startsols, vrblvl=vrblvl)
    if vrblvl > 0:
        for (idx, sol) in enumerate(newsols):
            print('Solution', idx+1, ':')
            print(sol)
    err = verify(circle, newsols, vrblvl-1)
    if vrblvl > 0:
        print('the error :', err)
    fail = int(err > 1.0e-8)
    sqrt5 = sqrt(5)
    sweepline = f'(y - {sqrt5:.15e})*(1-s) + y*s;'
    circle = ['x^2 + y^2 - 1;', sweepline]
    first = make_solution(['x', 'y', 's'], [complex(0,2), sqrt5, 0])
    second = make_solution(['x', 'y', 's'], [complex(0,-2), sqrt5, 0])
    startsols = [first, second]
    newsols = double_real_sweep(circle, startsols, vrblvl=vrblvl)
    if vrblvl > 0:
        for (idx, sol) in enumerate(newsols):
            print('Solution', idx+1, ':')
            print(sol)
    err = verify(circle, newsols, vrblvl-1)
    if vrblvl > 0:
        print('the error :', err)
    fail = fail + int(err > 1.0e-8)
    return fail

def test_double_double_real_sweep(vrblvl=0):
    """
    Runs a real sweep on two points on the unit circle: (1,0), (-1,0),
    moving the second coordinate from 0 to 2.
    The sweep will stop at the quadratic turning point: (0,1).
    We can also run the sweep starting at two complex points:
    (2*j, sqrt(5)) and (-2*j, sqrt(5)), moving the second coordinate
    from sqrt(5) to 0.  This sweep will also stop at (0,1).
    """
    if vrblvl > 0:
        print('in test_double_double_real_sweep ...')
    circle = ['x^2 + y^2 - 1;', 'y*(1-s) + (y-2)*s;']
    first = make_solution(['x', 'y', 's'], [1, 0, 0])
    second = make_solution(['x', 'y', 's'], [-1, 0, 0])
    startsols = [first, second]
    newsols = double_double_real_sweep(circle, startsols, vrblvl=vrblvl)
    if vrblvl > 0:
        for (idx, sol) in enumerate(newsols):
            print('Solution', idx+1, ':')
            print(sol)
    err = verify(circle, newsols, vrblvl-1)
    if vrblvl > 0:
        print('the error :', err)
    fail = int(err > 1.0e-8)
    sqrt5 = sqrt(5)
    sweepline = f'(y - {sqrt5:.15e})*(1-s) + y*s;'
    circle = ['x^2 + y^2 - 1;', sweepline]
    first = make_solution(['x', 'y', 's'], [complex(0,2), sqrt5, 0])
    second = make_solution(['x', 'y', 's'], [complex(0,-2), sqrt5, 0])
    startsols = [first, second]
    newsols = double_double_real_sweep(circle, startsols, vrblvl=vrblvl)
    if vrblvl > 0:
        for (idx, sol) in enumerate(newsols):
            print('Solution', idx+1, ':')
            print(sol)
    err = verify(circle, newsols, vrblvl-1)
    if vrblvl > 0:
        print('the error :', err)
    fail = fail + int(err > 1.0e-8)
    return fail

def test_quad_double_real_sweep(vrblvl=0):
    """
    Runs a real sweep on two points on the unit circle: (1,0), (-1,0),
    moving the second coordinate from 0 to 2.
    The sweep will stop at the quadratic turning point: (0,1).
    We can also run the sweep starting at two complex points:
    (2*j, sqrt(5)) and (-2*j, sqrt(5)), moving the second coordinate
    from sqrt(5) to 0.  This sweep will also stop at (0,1).
    """
    if vrblvl > 0:
        print('in test_quad_double_real_sweep ...')
    circle = ['x^2 + y^2 - 1;', 'y*(1-s) + (y-2)*s;']
    first = make_solution(['x', 'y', 's'], [1, 0, 0])
    second = make_solution(['x', 'y', 's'], [-1, 0, 0])
    startsols = [first, second]
    newsols = quad_double_real_sweep(circle, startsols, vrblvl=vrblvl)
    if vrblvl > 0:
        for (idx, sol) in enumerate(newsols):
            print('Solution', idx+1, ':')
            print(sol)
    err = verify(circle, newsols, vrblvl-1)
    if vrblvl > 0:
        print('the error :', err)
    fail = int(err > 1.0e-8)
    sqrt5 = sqrt(5)
    sweepline = f'(y - {sqrt5:.15e})*(1-s) + y*s;'
    circle = ['x^2 + y^2 - 1;', sweepline]
    first = make_solution(['x', 'y', 's'], [complex(0,2), sqrt5, 0])
    second = make_solution(['x', 'y', 's'], [complex(0,-2), sqrt5, 0])
    startsols = [first, second]
    newsols = quad_double_real_sweep(circle, startsols, vrblvl=vrblvl)
    if vrblvl > 0:
        for (idx, sol) in enumerate(newsols):
            print('Solution', idx+1, ':')
            print(sol)
    err = verify(circle, newsols, vrblvl-1)
    if vrblvl > 0:
        print('the error :', err)
    fail = fail + int(err > 1.0e-8)
    return fail

def main():
    """
    Runs some tests.
    """
    lvl = 1
    fail = test_double_complex_sweep(lvl)
    fail = test_double_double_complex_sweep(lvl)
    fail = test_quad_double_complex_sweep(lvl)
    fail = test_double_real_sweep(lvl)
    fail = test_double_double_real_sweep(lvl)
    fail = test_quad_double_real_sweep(lvl)
    if fail == 0:
        print('=> All tests passed.')
    else:
        print('Number of failed tests :', fail)

if __name__== '__main__':
    main()



"""
Backelin's Lemma states that the cyclic n-roots problem has an m-dimensional
solution set for n = L*m^2, where L is no multiple of k^2, m >= 2.
This script provides an exact representation of the solution set,
following the tropical formulation with m-1 free parameters.
The tropical version of Backelin's Lemma appeared in the proceedings of
Computer Algebra in Scientific Computing, 15th International Workshop, 
CASC 2013, Berlin, Germany, edited by Vladimir P.  Gerdt, Wolfram Koepf, 
Ernst W. Mayr, and Evgenii V. Vorozhtsov;
Lecture Notes in Computer Science, Volume 8136, pages 10-29, 2013;
in the paper Polyhedral Methods for Space Curves Exploiting Symmetry 
Applied to the Cyclic n-roots Problem, by Danko Adrovic and Jan Verschelde.
"""

def randcmplx():
    """
    Returns a random complex number on the unit circle.
    """
    from random import uniform
    from math import cos, sin, pi
    angle = uniform(0, 2*pi)
    return complex(cos(angle), sin(angle))

def strcmplx(nbr, imagunit, mulsym=''):
    """
    Returns the string representation of a complex number,
    where imagunit contains the string to use for the
    imaginary unit, and mulsym is the symbol for multiplication
    with the imaginary unit.
    """
    nre = '%.16e' % nbr.real
    nim = '%+.16e' % nbr.imag
    return '(' + nre + nim + mulsym + imagunit + ')'

def constants(m, L):
    """
    Returns the tuple of constants u and gamma
    for n = m*L**2.
    """
    from cmath import exp, pi
    alpha = m*(m*L - 1)
    beta = alpha % 2
    i = complex(0, 1)
    u = exp(i*2*pi/(m*L))
    gamma = exp(i*pi*beta/(m*L))
    return (u, gamma)

def component(m, L, u, gamma):
    """
    Returns a list of strings, where the k-th string
    stores the representation of the k-th coordinate.
    The m-1 parameters are t0, t1, .., tm-2.
    """
    result = []
    ustr = '(%.16e %+.16ej)' % (u.real, u.imag)
    gstr = '(%.16e %+.16ej)' % (gamma.real, gamma.imag)
    for k in range(L*m):
        s = '+' + ustr + '**' + str(k)
        for i in range(m-1):
            s = s + '*t' + str(i)
            result.append(s)
        s = '+' + gstr + '*' + ustr + '**' + str(k)
        for i in range(m-1):
            e = -m + 1 + i
            s = s + '*t' + str(i) + '**(' + str(e) + ')'
        result.append(s)
    return result

def component_monomials(m, L):
    """
    Returns a list of strings, where the k-th string
    stores the representation of the k-th coordinate.
    The m-1 parameters are t0, t1, .., tm-2.
    Only the monomials are returned, not the coefficients,
    which are not needed for the degree computation.
    """
    result = []
    for k in range(L*m):
        s = '+ 1'
        for i in range(m-1):
            s = s + '*t' + str(i)
            result.append(s)
        s = '+ 1'
        for i in range(m-1):
            e = -m + 1 + i
            s = s + '*t' + str(i) + '**(' + str(e) + ')'
        result.append(s)
    return result

def sample(m, roots):
    """
    Generates m-1 random complex numbers as parameters
    and evaluates the representation for the roots.
    """
    d = globals()
    for k in range(m-1):
        var = 't' + str(k)
        d[var] = randcmplx()
    result = []
    for r in roots:
        result.append(eval(r))
    return result

def evaluate(point):
    """
    Evaluates the point in the cyclic n-roots problem,
    where n = len(point).
    """
    from phcpy.families import cyclic
    n = len(point)
    f = cyclic(n)
    d = globals()
    for k in range(n):
        var = 'x' + str(k)
        d[var] = point[k]
    for k in range(n):
        var = 'x' + str(k)
        print var, '=', d[var]
    result = []
    for r in f:
        result.append(eval(r[0:-1]))
    return result

def embedpol(nbslk, pol):
    """
    Given in pol the string representation of a polynomial
    in several variables, the string pol is returned,
    augmented with a random linear sum of as many slack
    variables as the value of nbslk.
    """
    hyp = ""
    for k in range(1, nbslk+1):
        nbr = randcmplx()
        cff = strcmplx(nbr, 'i', '*')
        hyp = hyp + ' + ' + cff + '*zz' + str(k)
    result = pol[0:-1] + hyp + ';'
    return result

def embedsys(nbslk, sys, addslack=True):
    """
    Given on input in sys a list of string representations
    for polynomials, on return is the list of strings
    representing the embedding with as many slack variables
    as the value of nbslk, provided addslack equals True.
    Otherwise, the system on return is just a copy of sys.
    """
    result = []
    if addslack:
        for pol in sys:
            result.append(embedpol(nbslk, pol))
    else:
        for pol in sys:
            result.append(pol)
    return result

def hyperplane(dim, point, addslack=True):
    """
    Returns a random linear equation in len(point) + dim
    variables through the point.
    The first len(point) variables have name 'x' followed by
    an index which starts its count at zero.
    The dim slack variables have name 'zz' and their index
    count starts at one, provided addslack is True.
    The constant term of the hyperplane is computed so that
    the point evaluates to zero at the hyperplane.
    """
    nvr = len(point)
    hyp = ""    # to return in PHCpack format
    evahyp = "" # in Python format for evaluation
    for k in range(nvr):
        nbr = randcmplx()
        cff = strcmplx(nbr, 'i', '*')
        evacff = strcmplx(nbr, 'j')
        hyp = hyp + ' + ' + cff + '*x' + str(k)
        evahyp = evahyp + ' + ' + evacff + '*x' + str(k)
    if addslack:
        for k in range(1, dim+1):
            nbr = randcmplx()
            cff = strcmplx(nbr, 'i', '*')
            evacff = strcmplx(nbr, 'j')
            hyp = hyp + ' + ' + cff + '*zz' + str(k)
            evahyp = evahyp + ' + ' + evacff + '*zz' + str(k)
    d = globals()
    for k in range(nvr):
        var = 'x' + str(k)
        d[var] = point[k]
    if addslack:
        for k in range(1, dim+1):
            var = 'zz' + str(k)
            d[var] = 0.0
    cst = eval(evahyp)
    result = hyp + ' + (%.16e %+.16e*i)' % (-cst.real, -cst.imag) + ';'
    return result

def embedpoint(dim, point, addslack=True):
    """
    Embeds the point in a witness set representation
    for a cyclic n-roots solution set of dimension dim.
    If addslack, then slack variables will be added
    to square the embedded system.
    """
    from phcpy.families import cyclic
    nvr = len(point)
    sys = cyclic(nvr)
    emb = embedsys(dim, sys, addslack)
    for p in point:
        print p
    for k in range(dim):
        hyp = hyperplane(dim, point, addslack)
        emb.append(hyp)
    return emb

def embedtofile(embsys, point, addslack=True):
    """
    Prompts the user for a file name and writes the
    embedded system in embsys and the point to file.
    If addslack, then slack variables will be used
    to square the system.
    """
    name = raw_input('Give a file name : ')
    file = open(name, 'w')
    if addslack:
        file.write(str(len(embsys)) + '\n')
    else:
        file.write(str(len(embsys)) + ' ' + str(len(point))+ '\n')
    for pol in embsys:
        file.write(pol + '\n')
    from phcpy.solutions import make_solution
    vars = []  # symbols for the variables
    vals = []  # values for the variables
    for k in range(len(point)):
        vars.append('x' + str(k))
        vals.append(point[k])
    dim = len(embsys) - len(point)
    if addslack:
        for k in range(1, dim+1):
            vars.append('zz' + str(k))
            vals.append(complex(0))
    print 'vars = \n', vars
    print 'vals = \n', vals
    sol = make_solution(vars, vals)
    n = len(point)
    file.write('\nTITLE : sample of tropical Backelin cyclic %d-roots\n' % n)
    file.write('\nTHE SOLUTIONS :\n')
    file.write('1 ' + str(len(vars)) + '\n')
    file.write('====================================================\n')
    file.write('solution 1 :\n')
    file.write(sol + '\n')

def ask_inputs():
    """
    The components of cyclic n-roots according to Backelin's Lemma
    occur in dimension n = m**2*L.  This script prompts the user
    to enter m, L, and then return m, L, and n.
    """
    m = input("Give the m in n = m**2*L : ")
    L = input("Give the L in n = m**2*L : ")
    n = m**2*L
    return (m, L, n)

def sample_component():
    """
    Prompts the user for the parameters m and L.
    Computes one random point on the cyclic n-roots component,
    for n = m**2*L and embeds the sample point in a witness set.
    """
    (m, L, n) = ask_inputs()
    print 'The dimension n = %d**2*%d = %d.' % (m, L, n)
    u, gamma = constants(m, L)
    print 'u =', u
    print 'gamma =', gamma
    coords = component(m, L, u, gamma)
    print 'coordinates of a cyclic %d-roots component :' % n
    for c in coords:
        print c
    point = sample(m, coords)
    print 'a random point:'
    for p in point:
        print '%.16e %+.16ej)' % (p.real, p.imag)
    residual = evaluate(point)
    print 'the residual :'
    for r in residual:
        print r
    print 'the sum :', sum([abs(r) for r in residual])
    ans = raw_input('embed sample point in witness set ? (y/n) ')
    if(ans == 'y'):
        ans = raw_input('use slack variables to square the system ? (y/n) ')
        addslack = (ans == 'y')
        emb = embedpoint(m-1, point, addslack)
        print 'the embedded system :\n', emb
        ans = raw_input('write the witness set to file ? (y/n) ')
        if(ans == 'y'):
            embedtofile(emb, point, addslack)

def compute_degree():
    """
    Prompts the user for the parameters m and L.
    Computes the degree of the cyclic n-roots set,
    for n = m**2*L.
    """
    (m, L, n) = ask_inputs()
    print 'The dimension n = %d**2*%d = %d.' % (m, L, n)
    coords = component_monomials(m, L)
    print 'monomial coordinates of a cyclic %d-roots component :' % n
    for c in coords:
        print c
    dim = m-1
    print 'dimension =', dim
    pol = ''
    for c in coords:
        pol = pol + c
    print pol
    sys = []
    for k in range(dim):
        sys.append(pol + ';')
    print sys
    from phcpy.solver import mixed_volume
    deg = mixed_volume(sys)
    print 'the degree of cyclic %d-roots set : %d' % (n, deg)

def degree(m, L):
    """
    Returns the degree of the cyclic n-roots solution set,
    for n = m**2*L.
    """
    coords = component_monomials(m, L)
    dim = m-1
    pol = ''
    for c in coords:
        pol = pol + c
    sys = []
    for k in range(m-1):
        sys.append(pol + ';')
    from phcpy.solver import mixed_volume
    deg = mixed_volume(sys)
    return deg

def compute_degrees():
    """
    Computes the degrees of cyclic n-roots solution sets,
    for n = m**2*L, for a given range of m.
    """
    bench = [(4, 1), (4, 2), (4, 3), (8, 1), (4, 5), (4, 6), \
             (8, 2), (12, 1), (4, 10), (4, 11), (8, 3), (4, 13), \
             (4, 14), (4, 15), (16, 1), (4, 17), (12, 2), (4, 19), \
             (8, 5), (4, 21), (4, 22)]
    for (m, L) in bench:
        d = degree(m, L)
        n = m**2*L
        print 'degree of cyclic %d-roots set : %d, m = %d, L = %d' \
            % (n, d, m, L)

def main():
    """
    Presents the user with a menu of choices and 
    then calls the selected script.
    """
    print 'MENU for cyclic n-roots, n = m**2*L :'
    print '  1. sample a point and embed as witness set'
    print '  2. compute the degree of one cyclic n-roots set'
    print '  3. compute the degree of cyclic n-roots bench sets'
    answer = raw_input('Type 1, 2, or 3 to select : ')
    if(answer == '1'):
        sample_component()   
    elif(answer == '2'):
        compute_degree()
    elif(answer == '3'):
        compute_degrees()
    else:
        print 'invalid choice, please try again...'

main()



from phc_html import *
from phc_file import get_file_id, check_file_exist
from phc_status import PHC_Status

def PHC_Update(Uid,Pid,poly_id,phcdb, sol_page=0):
   from os import path
   if poly_id == '':
      Print_Poly_Empty(Uid)
   else:
      file_id = get_file_id(Uid, poly_id)
      if check_file_exist(file_id+".poly"):
         Print_Notice_Top()
         status = PHC_Status(Uid, poly_id, file_id, phcdb)
         Print_Area_Eq(file_id+".poly")
         if path.isfile(file_id+".err"):
            phcdb.userstatus(Uid, "NULL")
            Print_Button_Status(status, Uid, Pid, poly_id)
            Print_Notice_Mid(status, Uid, poly_id)

         elif status == 1: # being computed
            Print_Button_Status(status, Uid, Pid, poly_id)
            Print_Notice_Mid(status, Uid, poly_id)
            
         elif status == 0: # killed
            Print_Button_Status(status, Uid, Pid, poly_id)
            Print_Notice_Mid(status, Uid, poly_id)
            Print_Report(status, file_id, Uid, poly_id)
            
         else: # Just Solved
            phcdb.userstatus(Uid, "NULL")
            phcdb.slvpoly2(Uid, poly_id)
            Print_Button_Status(status, Uid, Pid, poly_id)
            Print_Notice_Mid(status, Uid, poly_id)
            Print_Report(status, file_id, Uid, poly_id, sol_page)

      else:
         Print_Poly_Empty(Uid)
         if poly_id == "current":
            phcdb.userstatus(Uid, "NULL")

def Print_Report(status, file_id, Uid, poly_id, sol_page):
   if status == 4 or status == 0:
      Print_Area_Report(file_id+".phc","report")
   elif status != 9:
      Print_Area_Sol(file_id+".sol",Uid, poly_id,"sols", sol_page)
   else:
      Print_Area_Report(file_id+".phc","report")



"""
We illustrate deflation on the example of Griewank and Osborne.
The point (0, 0) is an irregular singularity and Newton fails
to converge even if we start already quite close to (0, 0).
Deflation solves this problem.
"""
p = ['(29/16)*x^3 - 2*x*y;', 'x^2 - y;']
from phcpy.solutions import make_solution
sol = make_solution(['x', 'y'],[1.0e-6, 1.0e-6])
print 'the initial solution :'
print sol
from phcpy.solver import newton_step
from phcpy.solutions import diagnostics
sols = [sol]
for k in range(5):
    sols = newton_step(p, sols)
    (err, rco, res) = diagnostics(sols[0])
    print 'forward error :', err
    print 'estimate for inverse condition :', rco
    print 'backward error :', res
print 'the solution after five Newton steps :'
print sols[0]
from phcpy.solver import standard_deflate
sols = standard_deflate(p, sols)
print 'after deflation :'
print sols[0]
